import threading
import traceback
import collections
from werkzeug.exceptions import BadRequest
import importlib
from pkgutil import iter_modules
from chellow.models import (
    Session, Supply, BatchFile, BillType, Tpr, ReadType, Batch)
import chellow
import chellow.bill_parser_engie_xls
from chellow.utils import keydefaultdict, utc_datetime_now
from io import BytesIO


import_id = 0
import_lock = threading.Lock()
imports = {}


def find_parser_names():
    return [
        name[12:] for _, name, _ in iter_modules(chellow.__path__)
        if name.startswith('bill_parser_')
    ]


class BillImport(threading.Thread):
    def __init__(self, batch):
        threading.Thread.__init__(self)
        global import_id
        self.import_id = import_id
        import_id += 1

        self.batch_id = batch.id
        self.successful_bills = []
        self.failed_bills = []
        self.log = collections.deque()
        self.bill_num = None
        self.parser = None

    def _log(self, msg):
        with import_lock:
            self.log.appendleft(
                utc_datetime_now().strftime("%Y-%m-%d %H:%M:%S") + ' - ' + msg)

    def status(self):
        if self.isAlive():
            if self.bill_num is not None:
                return "Inserting raw bills: I've reached bill number " + \
                    str(self.bill_num) + "."
            elif self.parser is not None:
                return "Parsing file: I've reached line number " + \
                    str(self.parser.line_number) + "."
            else:
                return 'Running'
        else:
            return 'Not running'

    def run(self):
        sess = None
        try:
            sess = Session()
            batch = Batch.get_by_id(sess, self.batch_id)
            raw_bills = []
            for bf in sess.query(BatchFile).filter(
                    BatchFile.batch == batch).order_by(
                    BatchFile.upload_timestamp):

                self.parser = _process_batch_file(sess, bf, self._log)
                raw_bills.extend(self.parser.make_raw_bills())
                self._log(f"Successfully parsed the file {bf.filename}")

            bill_types = keydefaultdict(
                lambda k: BillType.get_by_code(sess, k))

            tprs = keydefaultdict(
                lambda k: None if k is None else Tpr.get_by_code(sess, k))

            read_types = keydefaultdict(
                lambda k: ReadType.get_by_code(sess, k))

            self._log(
                "Successfully parsed all the files, and now the raw bills "
                "will be inserted.")
            for self.bill_num, raw_bill in enumerate(raw_bills):
                try:
                    mpan_core = raw_bill['mpan_core']
                    supply = Supply.get_by_mpan_core(sess, mpan_core)
                    with sess.begin_nested():
                        bill = batch.insert_bill(
                            sess, raw_bill['account'], raw_bill['reference'],
                            raw_bill['issue_date'], raw_bill['start_date'],
                            raw_bill['finish_date'], raw_bill['kwh'],
                            raw_bill['net'], raw_bill['vat'],
                            raw_bill['gross'],
                            bill_types[raw_bill['bill_type_code']],
                            raw_bill['breakdown'], supply)
                        for raw_read in raw_bill['reads']:
                            bill.insert_read(
                                sess, tprs[raw_read['tpr_code']],
                                raw_read['coefficient'], raw_read['units'],
                                raw_read['msn'], raw_read['mpan'],
                                raw_read['prev_date'], raw_read['prev_value'],
                                read_types[raw_read['prev_type_code']],
                                raw_read['pres_date'], raw_read['pres_value'],
                                read_types[raw_read['pres_type_code']])
                        self.successful_bills.append(raw_bill)
                except BadRequest as e:
                    raw_bill['error'] = str(e.description)
                    self.failed_bills.append(raw_bill)

            if len(self.failed_bills) == 0:
                sess.commit()
                self._log(
                    "All the bills have been successfully loaded and attached "
                    "to the batch.")
            else:
                sess.rollback()
                self._log(
                    "The import has finished, but there were " +
                    str(len(self.failed_bills)) + " failures, and so the "
                    "whole import has been rolled back.")

        except BadRequest as e:
            sess.rollback()
            self._log("Problem: " + e.description)
        except BaseException:
            sess.rollback()
            self._log("I've encountered a problem: " + traceback.format_exc())
        finally:
            if sess is not None:
                sess.close()

    def make_fields(self):
        with import_lock:
            fields = {
                'log': tuple(self.log), 'is_alive': self.isAlive(),
                'importer_id': self.import_id}
            if not self.isAlive():
                fields['successful_bills'] = self.successful_bills
                fields['failed_bills'] = self.failed_bills
            return fields


def _process_batch_file(sess, batch_file, log_f):
    data = batch_file.data
    parser_name = batch_file.parser_name

    if len(data) == 0:
        raise BadRequest("File has zero length")

    try:
        imp_mod = importlib.import_module(
            f'chellow.bill_parser_{parser_name}')
    except ImportError:
        raise BadRequest(
            f"Can't find a parser with the name '{parser_name}'.")

    parser = imp_mod.Parser(BytesIO(data))
    log_f(f"Starting to parse the file with '{parser_name}'.")

    return parser


def start_bill_import(batch):
    with import_lock:
        bi = BillImport(batch)
        imports[bi.import_id] = bi
        bi.start()

    return bi.import_id


def get_bill_import_ids(batch):
    with import_lock:
        return [k for k, v in imports.items() if v.batch_id == batch.id]


def get_bill_import(id):
    with import_lock:
        return imports[id]



from chellow.utils import get_file_rates


def ccl(data_source):
    cache = data_source.caches

    for hh in data_source.hh_data:
        rates = get_file_rates(cache, 'ccl', hh['start-date'])
        rate = float(rates['ccl_gbp_per_msp_kwh'])
        hh['ccl-kwh'] = hh['msp-kwh']
        hh['ccl-rate'] = rate
        hh['ccl-gbp'] = hh['msp-kwh'] * rate



from sqlalchemy import or_
from sqlalchemy.sql.expression import null, true
import traceback
from chellow.models import (
    Session, Supply, Era, Site, SiteEra, Tpr, MeasurementRequirement, Ssc)
from chellow.utils import (
    hh_min, hh_max, hh_format, req_int, req_date, csv_make_val)
from chellow.views import chellow_redirect
import chellow.computer
from werkzeug.exceptions import BadRequest
import csv
import os
from flask import g
import threading


def content(supply_id, file_name, start_date, finish_date, user):
    caches = {}
    sess = None
    try:
        sess = Session()
        running_name, finished_name = chellow.dloads.make_names(
            'supply_virtual_bills_' + str(supply_id) + '.csv', user)
        f = open(running_name, mode='w', newline='')
        writer = csv.writer(f, lineterminator='\n')

        supply = Supply.get_by_id(sess, supply_id)

        forecast_date = chellow.computer.forecast_date()

        prev_titles = None

        for era in sess.query(Era).filter(
                Era.supply == supply, Era.start_date < finish_date, or_(
                    Era.finish_date == null(),
                    Era.finish_date > start_date)).order_by(Era.start_date):

            chunk_start = hh_max(era.start_date, start_date)
            chunk_finish = hh_min(era.finish_date, finish_date)
            site = sess.query(Site).join(SiteEra).filter(
                SiteEra.era == era, SiteEra.is_physical == true()).one()

            ds = chellow.computer.SupplySource(
                sess, chunk_start, chunk_finish, forecast_date, era,
                era.imp_supplier_contract is not None, caches)

            titles = [
                'Imp MPAN Core', 'Exp MPAN Core', 'Site Code', 'Site Name',
                'Account', 'From', 'To', ''
            ]

            output_line = [
                era.imp_mpan_core, era.exp_mpan_core, site.code,
                site.name, ds.supplier_account, hh_format(ds.start_date),
                hh_format(ds.finish_date), '']

            mop_titles = ds.contract_func(
                era.mop_contract, 'virtual_bill_titles')()
            titles.extend(['mop-' + t for t in mop_titles])

            ds.contract_func(era.mop_contract, 'virtual_bill')(ds)
            bill = ds.mop_bill
            for title in mop_titles:
                if title in bill:
                    output_line.append(bill[title])
                    del bill[title]
                else:
                    output_line.append('')

            for k in sorted(bill.keys()):
                output_line.extend([k, bill[k]])

            output_line.append('')
            dc_titles = ds.contract_func(
                era.dc_contract, 'virtual_bill_titles')()
            titles.append('')
            titles.extend(['dc-' + t for t in dc_titles])

            ds.contract_func(era.dc_contract, 'virtual_bill')(ds)
            bill = ds.dc_bill
            for title in dc_titles:
                output_line.append(bill.get(title, ''))
                if title in bill:
                    del bill[title]
            for k in sorted(bill.keys()):
                output_line.extend([k, bill[k]])

            tpr_query = sess.query(Tpr).join(MeasurementRequirement). \
                join(Ssc).join(Era).filter(
                    Era.start_date <= chunk_finish, or_(
                        Era.finish_date == null(),
                        Era.finish_date >= chunk_start)
                ).order_by(Tpr.code).distinct()

            if era.imp_supplier_contract is not None:
                output_line.append('')
                supplier_titles = ds.contract_func(
                    era.imp_supplier_contract, 'virtual_bill_titles')()
                for tpr in tpr_query.filter(
                        Era.imp_supplier_contract != null()):
                    for suffix in ('-kwh', '-rate', '-gbp'):
                        supplier_titles.append(tpr.code + suffix)
                titles.append('')
                titles.extend(['imp-supplier-' + t for t in supplier_titles])

                ds.contract_func(era.imp_supplier_contract, 'virtual_bill')(ds)
                bill = ds.supplier_bill

                for title in supplier_titles:
                    if title in bill:
                        output_line.append(bill[title])
                        del bill[title]
                    else:
                        output_line.append('')

                for k in sorted(bill.keys()):
                    output_line.extend([k, bill[k]])

            if era.exp_supplier_contract is not None:
                ds = chellow.computer.SupplySource(
                    sess, chunk_start, chunk_finish, forecast_date, era, False,
                    caches)

                output_line.append('')
                supplier_titles = ds.contract_func(
                    era.exp_supplier_contract, 'virtual_bill_titles')()
                for tpr in tpr_query.filter(
                        Era.exp_supplier_contract != null()):
                    for suffix in ('-kwh', '-rate', '-gbp'):
                        supplier_titles.append(tpr.code + suffix)
                titles.append('')
                titles.extend(['exp-supplier-' + t for t in supplier_titles])

                ds.contract_func(
                    era.exp_supplier_contract, 'virtual_bill')(ds)
                bill = ds.supplier_bill
                for title in supplier_titles:
                    output_line.append(bill.get(title, ''))
                    if title in bill:
                        del bill[title]

                for k in sorted(bill.keys()):
                    output_line.extend([k, bill[k]])

            if titles != prev_titles:
                prev_titles = titles
                writer.writerow([str(v) for v in titles])
            for i, val in enumerate(output_line):
                output_line[i] = csv_make_val(val)
            writer.writerow(output_line)
    except BadRequest as e:
        writer.writerow(["Problem: " + e.description])
    except BaseException:
        writer.writerow([traceback.format_exc()])
    finally:
        if sess is not None:
            sess.close()
        if f is not None:
            f.close()
            os.rename(running_name, finished_name)


def do_get(sess):
    supply_id = req_int('supply_id')
    file_name = 'supply_virtual_bills_' + str(supply_id) + '.csv'
    start_date = req_date('start')
    finish_date = req_date('finish')
    args = (supply_id, file_name, start_date, finish_date, g.user)

    threading.Thread(target=content, args=args).start()
    return chellow_redirect("/downloads", 303)



import traceback
from chellow.utils import req_date, req_int, hh_format
from chellow.models import Site, Session, Era, SiteEra, Supply
from flask import request, g
import csv
import chellow.dloads
from chellow.views import chellow_redirect
import threading
import sys
import os
from sqlalchemy import true, or_, null
from sqlalchemy.orm import joinedload


METER_ORDER = {'hh': 0, 'amr': 1, 'nhh': 2, 'unmetered': 3, '': 4}


def content(start_date, finish_date, site_id, user):
    sess = None
    try:
        sess = Session()
        running_name, finished_name = chellow.dloads.make_names(
            'sites_duration.csv', user)
        f = open(running_name, mode='w', newline='')
        writer = csv.writer(f, lineterminator='\n')
        writer.writerow(
            (
                "Site Id", "Site Name", "Associated Site Ids", "Sources",
                "Generator Types", "From", "To", "Imported kWh",
                "Displaced kWh", "Exported kWh", "Used kWh", "Parasitic kWh",
                "Generated kWh", "Meter Type"))

        streams = (
            'imp_net', 'displaced', 'exp_net', 'used', 'exp_gen', 'imp_gen')

        sites = sess.query(Site).order_by(Site.code)
        if site_id is not None:
            sites = sites.filter(Site.id == site_id)

        start_date_str = hh_format(start_date)
        finish_date_str = hh_format(finish_date)

        for site in sites:
            assoc = ' '.join(
                s.code for s in site.find_linked_sites(
                    sess, start_date, finish_date))

            totals = dict((stream, 0) for stream in streams)

            metering_type = ''
            source_codes = set()
            gen_types = set()
            for era in sess.query(Era).join(SiteEra).filter(
                    SiteEra.is_physical == true(), SiteEra.site == site,
                    Era.start_date <= finish_date, or_(
                        Era.finish_date == null(),
                        Era.finish_date >= start_date)).distinct().options(
                            joinedload(Era.supply).joinedload(Supply.source),
                            joinedload(Era.supply).
                            joinedload(Supply.generator_type)):
                supply = era.supply
                source_codes.add(supply.source.code)
                gen_type = supply.generator_type
                if gen_type is not None:
                    gen_types.add(gen_type.code)
                era_meter_type = era.meter_category
                if METER_ORDER[era_meter_type] < METER_ORDER[metering_type]:
                    metering_type = era_meter_type

            assoc_str = ','.join(sorted(assoc))
            sources_str = ','.join(sorted(source_codes))
            generators_str = ','.join(sorted(gen_types))

            for hh in site.hh_data(sess, start_date, finish_date):
                for stream in streams:
                    totals[stream] += hh[stream]

            writer.writerow(
                (
                    site.code, site.name, assoc_str, sources_str,
                    generators_str, start_date_str, finish_date_str,
                    totals['imp_net'], totals['displaced'], totals['exp_net'],
                    totals['used'], totals['exp_gen'], totals['imp_gen'],
                    metering_type))

            # Prevent long-running transaction
            sess.rollback()
    except BaseException:
        msg = traceback.format_exc()
        sys.stderr.write(msg)
        writer.writerow([msg])
    finally:
        if sess is not None:
            sess.close()
        if f is not None:
            f.close()
            os.rename(running_name, finished_name)


def do_get(sess):
    start_date = req_date('start')
    finish_date = req_date('finish')
    site_id = req_int('site_id') if 'site_id' in request.values else None
    args = (start_date, finish_date, site_id, g.user)
    threading.Thread(target=content, args=args).start()
    return chellow_redirect("/downloads", 303)



from collections import defaultdict
from datetime import datetime as Datetime
from functools import lru_cache
from itertools import combinations, count
from math import log10
from types import MappingProxyType

import chellow.bank_holidays
import chellow.utils
from chellow.models import (
    Bill, BillType, Channel, ClockInterval, Contract, Era, HhDatum, Llfc,
    MeasurementRequirement, Mtc, Party, RateScript, ReadType, RegisterRead,
    SiteEra, Source, Ssc, Supply, Tpr, hh_after, hh_before)
from chellow.utils import (
    HH, PropDict, YEAR, c_months_u, ct_datetime, ct_datetime_now, hh_format,
    hh_max, hh_min, hh_range, loads, to_ct, to_tz, to_utc, utc_datetime,
    utc_datetime_now)

from dateutil.relativedelta import relativedelta

from pytz import timezone, utc

from sqlalchemy import Float, cast, or_
from sqlalchemy.orm import aliased, joinedload
from sqlalchemy.sql.expression import false, null

from werkzeug.exceptions import BadRequest

from zish import dumps


cons_types = ['construction', 'commissioning', 'operation']
lec_cats = list(
    (v + '-kwh', 'hist-' + v + '-kwh') for v in [
        'import-net', 'export-net', 'import-gen', 'export-gen',
        'import-3rd-party', 'export-3rd-party'])


@lru_cache()
def get_times(start_date, finish_date, forecast_date):
    if start_date > finish_date:
        raise BadRequest('The start date is after the finish date.')
    hist_start = start_date
    hist_finish = finish_date
    years_back = 0
    while hist_finish > forecast_date:
        hist_start -= relativedelta(years=1)
        hist_finish -= relativedelta(years=1)
        years_back += 1

    return MappingProxyType(
        {
            'history-finish': hist_finish,
            'history-start': hist_start,
            'years-back': years_back})


def contract_func(caches, contract, func_name):
    try:
        ns = caches['computer']['funcs'][contract.id]
    except KeyError:
        try:
            ccache = caches['computer']
        except KeyError:
            ccache = caches['computer'] = {}

        try:
            contr_func_cache = ccache['funcs']
        except KeyError:
            contr_func_cache = ccache['funcs'] = {}

        try:
            ns = contr_func_cache[contract.id]
        except KeyError:
            ns = {
                'db_id': contract.id, 'properties': contract.make_properties()}
            exec(contract.charge_script, ns)
            contr_func_cache[contract.id] = ns

    return ns.get(func_name, None)


def hh_rate(sess, caches, contract_id, date):
    try:
        return caches['computer']['rates'][contract_id][date]
    except KeyError:
        try:
            ccache = caches['computer']
        except KeyError:
            ccache = caches['computer'] = {}

        try:
            rss_cache = ccache['rates']
        except KeyError:
            rss_cache = ccache['rates'] = {}

        try:
            cont_cache = rss_cache[contract_id]
        except KeyError:
            cont_cache = rss_cache[contract_id] = {}

        try:
            return cont_cache[date]
        except KeyError:
            year_after = date + YEAR
            year_before = date - YEAR

            rs = sess.query(RateScript).filter(
                RateScript.contract_id == contract_id,
                RateScript.start_date <= date, or_(
                    RateScript.finish_date == null(),
                    RateScript.finish_date >= date)).first()

            if rs is None:
                rs = sess.query(RateScript).filter(
                    RateScript.contract_id == contract_id).order_by(
                    RateScript.start_date.desc()).first()
                if date < rs.start_date:
                    cstart = year_before
                    cfinish = min(year_after, rs.start_date - HH)
                else:
                    cstart = max(rs.finish_date + HH, year_before)
                    cfinish = year_after
            else:
                cstart = max(rs.start_date, year_before)
                if rs.finish_date is None:
                    cfinish = year_after
                else:
                    cfinish = min(rs.finish_date, year_after)

            market_role_code = rs.contract.market_role.code
            if market_role_code == "M":
                seg = 'mop_rate_scripts/'
            elif market_role_code == "C":
                seg = 'dc_rate_scripts/'
            elif market_role_code == "X":
                seg = 'supplier_rate_scripts/'
            elif market_role_code == "Z":
                seg = 'non_core_rate_scripts/'
            else:
                raise Exception(
                    "The market role code " + market_role_code +
                    " isn't recognized.")

            vals = PropDict(
                f"the rate script {chellow.utils.url_root}{seg}{rs.id} ",
                loads(rs.script), [])
            for dt in hh_range(caches, cstart, cfinish):
                if dt not in cont_cache:
                    cont_cache[dt] = vals

            return vals


def forecast_date():
    now = ct_datetime_now()
    return to_utc(ct_datetime(now.year, now.month, 1))


def displaced_era(
        sess, caches, site, start_date, finish_date, forecast_date,
        has_scenario_generation=False):
    start_date_ct = to_ct(start_date)
    finish_date_ct = to_ct(finish_date)
    if (start_date_ct.year, start_date_ct.month) != \
            (finish_date_ct.year, finish_date_ct.month):
        raise BadRequest(
            "The start and end dates of a displaced period must be within the "
            "same month")
    t = get_times(start_date, finish_date, forecast_date)
    hs = to_ct(t['history-start'])
    month_start, month_finish = next(
        c_months_u(start_year=hs.year, start_month=hs.month))
    has_displaced = False
    eras = {}
    for site_era in sess.query(SiteEra).join(Era).join(Supply).join(Source). \
            filter(
                SiteEra.site == site, Era.start_date <= month_finish, or_(
                    Era.finish_date == null(),
                    Era.finish_date >= month_start)).options(
                        joinedload(SiteEra.era).joinedload(Era.supply).
                        joinedload(Supply.source)):
        era = site_era.era
        source_code = era.supply.source.code
        if site_era.is_physical and (
                source_code in ('gen', 'gen-net') or (
                    source_code == 'net' and sess.query(Channel).filter(
                        Channel.era == era, Channel.imp_related == false(),
                        Channel.channel_type == 'ACTIVE').first()
                    is not None)):
            has_displaced = True

        if source_code in ('net', 'gen-net') and era.imp_mpan_core is not None:
            eras[
                '_'.join(
                    (
                        era.pc.code, str(era.exp_mpan_core is None),
                        hh_format(era.start_date), era.imp_mpan_core))] = era

    if (has_displaced or has_scenario_generation) and len(eras) > 0:
        era = eras[sorted(eras.keys())[0]]
    else:
        era = None
    return era


def _tpr_dict(sess, caches, tpr_code):
    try:
        return caches['computer']['tprs'][tpr_code]
    except KeyError:
        try:
            ccache = caches['computer']
        except KeyError:
            ccache = caches['computer'] = {}

        try:
            tpr_cache = ccache['tprs']
        except KeyError:
            tpr_cache = ccache['tprs'] = {}

        tpr = Tpr.get_by_code(sess, tpr_code)
        days_of_week = dict([i, []] for i in range(7))
        tpr_dict = {
            'days-of-week': days_of_week, 'is-gmt': tpr.is_gmt,
            'datum-cache': {}}
        if tpr.is_teleswitch:
            contract = Contract.find_supplier_by_name(sess, 'teleswitch')
            if contract is None:
                cis = []
            else:
                tprs = hh_rate(
                    sess, caches, contract.id, utc_datetime_now())['tprs']
                try:
                    cis = tprs[tpr_code]
                except KeyError:
                    raise BadRequest(
                        "Can't find the TPR " + tpr_code +
                        " in the rate script of the 'teleswitch' supplier "
                        "contract.")
        else:
            cis = [
                {
                    'day_of_week': ci.day_of_week,
                    'start_month': ci.start_month,
                    'start_day': ci.start_day,
                    'start_hour': ci.start_hour,
                    'start_minute': ci.start_minute,
                    'end_month': ci.end_month,
                    'end_day': ci.end_day,
                    'end_hour': ci.end_hour,
                    'end_minute': ci.end_minute}
                for ci in sess.query(ClockInterval).filter(
                    ClockInterval.tpr == tpr)]

        for ci in cis:
            days_of_week[ci['day_of_week'] - 1].append(
                {
                    'start-month': ci['start_month'] * 100 + ci['start_day'],
                    'start-hour': ci['start_hour'] + ci['start_minute'] / 60,
                    'end-month': ci['end_month'] * 100 + ci['end_day'],
                    'end-hour': ci['end_hour'] + ci['end_minute'] / 60})

        tpr_cache[tpr_code] = tpr_dict
        return tpr_dict


def is_tpr(sess, caches, tpr_code, hh_date):
    try:
        return _tpr_dict(sess, caches, tpr_code)['datum-cache'][hh_date]
    except KeyError:
        tpr_dict = _tpr_dict(sess, caches, tpr_code)
        datum_cache = tpr_dict['datum-cache']

        days_of_week = tpr_dict['days-of-week']
        datum = datum_range(sess, caches, 0, hh_date, hh_date)[0]
        pref = 'utc-' if tpr_dict['is-gmt'] else 'ct-'

        decimal_hour = datum[pref + 'decimal-hour']
        fractional_month = datum[pref + 'month'] * 100 + \
            datum[pref + 'day']

        result = False

        for ci in days_of_week[datum[pref + 'day-of-week']]:
            if (
                    (
                        ci['start-hour'] < ci['end-hour'] and
                        ci['start-hour'] <= decimal_hour < ci['end-hour']
                    ) or (
                        ci['start-hour'] >= ci['end-hour'] and (
                            ci['start-hour'] <= decimal_hour or
                            decimal_hour < ci['end-hour']))) and \
                    ci['start-month'] <= fractional_month \
                    <= ci['end-month']:
                result = True
                break
        datum_cache[hh_date] = result
        return result


def cache_level(cache, key):
    try:
        return cache[key]
    except BaseException:
        cache[key] = new_cache = {}
        return new_cache


def datum_range(sess, caches, years_back, start_date, finish_date):
    try:
        return caches['computer']['datum_range'][years_back][start_date][
            finish_date]
    except KeyError:
        computer_cache = cache_level(caches, 'computer')
        range_cache = cache_level(computer_cache, 'datum_range')
        yb_range_cache = cache_level(range_cache, years_back)
        start_range_cache = cache_level(yb_range_cache, start_date)
        datum_cache = cache_level(computer_cache, 'datum')
        yb_datum_cache = cache_level(datum_cache, years_back)

        datum_list = []
        bh_cont = Contract.get_non_core_by_name(sess, 'bank_holidays')
        for dt in hh_range(caches, start_date, finish_date):
            try:
                datum = yb_datum_cache[dt]
            except KeyError:
                hist_date = dt - relativedelta(years=years_back)
                ct_dt = to_ct(dt)

                utc_is_month_end = (dt + HH).day == 1 and dt.day != 1
                ct_is_month_end = (ct_dt + HH).day == 1 and ct_dt.day != 1

                utc_decimal_hour = dt.hour + dt.minute / 60
                ct_decimal_hour = ct_dt.hour + ct_dt.minute / 60

                bhs = hh_rate(sess, caches, bh_cont.id, dt)['bank_holidays']

                bank_holidays = [b[5:] for b in bhs]
                utc_is_bank_holiday = dt.strftime("%m-%d") in bank_holidays
                ct_is_bank_holiday = ct_dt.strftime("%m-%d") in bank_holidays

                datum = MappingProxyType(
                    {
                        'hist-start': hist_date, 'start-date': dt,
                        'ct-day': ct_dt.day, 'utc-month': dt.month,
                        'utc-day': dt.day,
                        'utc-decimal-hour': utc_decimal_hour,
                        'utc-year': dt.year, 'utc-hour': dt.hour,
                        'utc-minute': dt.minute, 'ct-year': ct_dt.year,
                        'ct-month': ct_dt.month,
                        'ct-decimal-hour': ct_decimal_hour,
                        'ct-day-of-week': ct_dt.weekday(),
                        'utc-day-of-week': dt.weekday(),
                        'utc-is-bank-holiday': utc_is_bank_holiday,
                        'ct-is-bank-holiday': ct_is_bank_holiday,
                        'utc-is-month-end': utc_is_month_end,
                        'ct-is-month-end': ct_is_month_end,
                        'status': 'X', 'imp-msp-kvarh': 0,
                        'imp-msp-kvar': 0, 'exp-msp-kvarh': 0,
                        'exp-msp-kvar': 0, 'msp-kw': 0, 'msp-kwh': 0,
                        'hist-import-net-kvarh': 0,
                        'hist-export-net-kvarh': 0,
                        'anti-msp-kwh': 0, 'anti-msp-kw': 0,
                        'imp-msp-kvarh': 0, 'exp-msp-kvarh': 0,
                        'imp-msp-kvar': 0, 'exp-msp-kvar': 0,
                        'hist-imp-msp-kvarh': 0, 'hist-kwh': 0})
                yb_datum_cache[dt] = datum
            datum_list.append(datum)
        datum_tuple = start_range_cache[finish_date] = tuple(datum_list)
        return datum_tuple


class DataSource():
    def __init__(
            self, sess, start_date, finish_date, forecast_date, caches,
            era_maps, deltas=None):
        self.full_channels = True
        self.sess = sess
        self.caches = caches
        self.forecast_date = forecast_date
        self.start_date = start_date
        self.finish_date = finish_date
        self.deltas = deltas
        times = get_times(start_date, finish_date, forecast_date)
        self.years_back = times['years-back']
        self.history_start = times['history-start']
        self.history_finish = times['history-finish']

        self.is_green = False
        self.supplier_bill = defaultdict(int, {'problem': ''})
        self.mop_bill = defaultdict(int, {'problem': ''})
        self.dc_bill = defaultdict(int, {'problem': ''})
        self.hh_data = []
        self.supplier_bill_hhs = {}
        self.mop_bill_hhs = {}
        self.dc_bill_hhs = {}

        self.era_maps = {} if era_maps is None else era_maps
        era_map = {}
        for em_start, em in sorted(self.era_maps.items()):
            if em_start <= start_date:
                era_map = em
                break
        self.era_map = PropDict("scenario properties", era_map)
        self.era_map_llfcs = self.era_map.get('llfcs', {})
        self.era_map_pcs = self.era_map.get('pcs', {})
        self.era_map_sscs = self.era_map.get('sscs', {})
        self.era_map_supplier_contracts = self.era_map.get(
            'supplier_contracts', {})
        self.era_map_dc_contracts = self.era_map.get('dc_contracts', {})
        self.era_map_mop_contracts = self.era_map.get('mop_contracts', {})
        self.era_map_cops = self.era_map.get('cops', {})
        self.era_map_mpan_cores = self.era_map.get('mpan_cores', {})
        self.era_map_dnos = self.era_map.get('dnos', {})
        self.era_map_gsp_groups = self.era_map.get('gsp_groups', {})

    def contract_func(self, contract, func_name):
        return contract_func(self.caches, contract, func_name)

    def hh_rate(self, contract_id, date):
        try:
            return self.rate_cache[contract_id][date]
        except KeyError:
            return hh_rate(self.sess, self.caches, contract_id, date)
        except AttributeError:
            val = hh_rate(self.sess, self.caches, contract_id, date)
            self.rate_cache = self.caches['computer']['rates']
            return val

    def _add_problem(self, problem):
        self.supplier_bill['problem'] += problem
        self.mop_bill['problem'] += problem
        self.dc_bill['problem'] += problem


class SiteSource(DataSource):
    def __init__(
            self, sess, site, start_date, finish_date, forecast_date, caches,
            era=None, era_maps=None, deltas=None):
        DataSource.__init__(
            self, sess, start_date, finish_date, forecast_date, caches,
            era_maps, deltas=deltas)
        self.site = site
        self.bill = None
        self.stream_focus = 'gen-used'
        if era is None:
            self.is_displaced = False
        else:
            self.era = era
            self.is_displaced = True
            self.supply = self.era.supply
            self.mpan_core = era.imp_mpan_core
            self.dno = self.supply.dno
            self.dno_code = self.dno.dno_code

            if era.imp_llfc.code in self.era_map_llfcs:
                self.llfc_code = self.era_map_llfcs[era.imp_llfc.code]
                self.llfc = self.dno.get_llfc_by_code(sess, self.llfc_code)
            else:
                self.llfc = era.imp_llfc
                self.llfc_code = self.llfc.code

            if era.pc.code in self.era_map_pcs:
                self.pc_code = self.era_map_pcs[era.pc.code]
            else:
                self.pc_code = era.pc.code

            era_ssc_code = None if era.ssc is None else era.ssc.code
            if era_ssc_code in self.era_map_sscs:
                self.ssc_code = self.era_map_sscs[era_ssc_code]
            else:
                self.ssc_code = era_ssc_code

            if era.cop.code in self.era_map_cops:
                self.cop_code = self.era_map_cops[era.cop.code]
            else:
                self.cop_code = era.cop.code

            if era.imp_supplier_contract.id in self.era_map_supplier_contracts:
                self.supplier_contract = Contract.get_supplier_by_id(
                    sess,
                    self.era_map_supplier_contracts[
                        era.imp_supplier_contract.id])
            else:
                self.supplier_contract = era.imp_supplier_contract

            if era.dc_contract.id in self.era_map_dc_contracts:
                self.dc_contract = Contract.get_dc_by_id(
                    sess, self.era_map_dc_contracts[era.dc_contract.id])
            else:
                self.dc_contract = era.dc_contract

            if era.mop_contract.id in self.era_map_mop_contracts:
                self.mop_contract = Contract.get_mop_by_id(
                    sess, self.era_map_mop_contracts[era.mop_contract.id])
            else:
                self.mop_contract = era.mop_contract

            if 'sc' in self.era_map:
                self.sc = self.era_map['sc']
            else:
                self.sc = era.imp_sc

            self.properties = self.era_map.get('properties_overwritten', {})
            self.properties.update(era.props)
            self.properties.update(
                self.era_map.get('properties_overwrite', {}))

            self.is_import = True
            self.voltage_level_code = self.llfc.voltage_level.code
            self.is_substation = self.llfc.is_substation
            self.gsp_group_code = self.supply.gsp_group.code
            self.ssc = self.era.ssc
            self.ssc_code = None if self.ssc is None else self.ssc.code

        self.era_ids = set(
            s.id for s in sess.query(Era).join(SiteEra).join(Supply).
            join(Source).filter(
                SiteEra.site == site, Era.start_date <= self.history_finish,
                SiteEra.is_physical, Source.code != 'sub', or_(
                    Era.finish_date == null(),
                    Era.finish_date >= self.history_start)).all())
        if len(self.era_ids) == 0:
            rs = iter([])
        else:
            rs = iter(
                sess.query(
                    cast(HhDatum.value, Float), HhDatum.start_date,
                    HhDatum.status, Channel.imp_related,
                    Source.code).join(Channel).join(Era).join(Supply).join(
                    Source).filter(
                    Channel.channel_type == 'ACTIVE',
                    HhDatum.start_date >= self.history_start,
                    HhDatum.start_date <= self.history_finish,
                    Era.id.in_(list(self.era_ids))).order_by(
                        HhDatum.start_date))

        hh_value, hh_start_date, status, imp_related, source_code = next(
            rs, (None, None, None, None, None))
        hist_map = {}

        for hist_date in hh_range(
                self.caches, self.history_start, self.history_finish):
            export_net_kwh = 0
            import_net_kwh = 0
            export_gen_kwh = 0
            import_gen_kwh = 0
            import_3rd_party_kwh = 0
            statuses = set()
            export_3rd_party_kwh = 0
            while hh_start_date == hist_date:
                statuses.add(status)
                if not imp_related and source_code in ('net', 'gen-net'):
                    export_net_kwh += hh_value
                if imp_related and source_code in ('net', 'gen-net'):
                    import_net_kwh += hh_value
                if (imp_related and source_code == 'gen') or \
                        (not imp_related and source_code == 'gen-net'):
                    import_gen_kwh += hh_value
                if (not imp_related and source_code == 'gen') or \
                        (imp_related and source_code == 'gen-net'):
                    export_gen_kwh += hh_value
                if (imp_related and source_code == '3rd-party') or (
                        not imp_related and
                        source_code == '3rd-party-reverse'):
                    import_3rd_party_kwh += hh_value
                if (not imp_related and source_code == '3rd-party') or (
                        imp_related and
                        source_code == '3rd-party-reverse'):
                    export_3rd_party_kwh += hh_value

                hh_value, hh_start_date, status, imp_related, source_code = \
                    next(rs, (None, None, None, None, None))

            if len(statuses) == 1 and statuses.pop() == 'A':
                status = 'A'
            else:
                status = 'E'

            hh_values = {
                'status': status, 'hist-import-net-kwh': import_net_kwh,
                'hist-export-net-kwh': export_net_kwh,
                'hist-import-gen-kwh': import_gen_kwh,
                'hist-export-gen-kwh': export_gen_kwh,
                'hist-import-3rd-party-kwh': import_3rd_party_kwh,
                'hist-export-3rd-party-kwh': export_3rd_party_kwh,
                'hist-used-3rd-party-kwh':
                    import_3rd_party_kwh - export_3rd_party_kwh}

            hh_values['used-3rd-party-kwh'] = \
                hh_values['hist-used-3rd-party-kwh']
            hh_values['hist-kwh'] = hh_values['hist-used-gen-msp-kwh'] = \
                hh_values['hist-import-gen-kwh'] - \
                hh_values['hist-export-gen-kwh'] - \
                hh_values['hist-export-net-kwh']

            hh_values['msp-kwh'] = hh_values['used-gen-msp-kwh'] \
                = hh_values['hist-used-gen-msp-kwh']
            for cat_kwh, hist_cat_kwh in lec_cats:
                hh_values[cat_kwh] = hh_values[hist_cat_kwh]

            hh_values['hist-used-kwh'] = \
                hh_values['hist-used-gen-msp-kwh'] + \
                hh_values['hist-import-net-kwh'] + \
                hh_values['hist-used-3rd-party-kwh']

            hh_values['used-kwh'] = hh_values['hist-used-kwh']
            hh_values['import-net-kwh'] = hh_values['hist-import-net-kwh']
            hh_values['msp-kw'] = hh_values['used-gen-msp-kw'] = \
                hh_values['used-gen-msp-kwh'] * 2

            hist_map[hist_date] = hh_values

        for dtm in datum_range(
                sess, self.caches, self.years_back, start_date, finish_date):
            datum = dtm.copy()
            datum.update(hist_map[datum['hist-start']])
            self.hh_data.append(datum)
            self.supplier_bill_hhs[dtm['start-date']] = {}
            self.mop_bill_hhs[dtm['start-date']] = {}
            self.dc_bill_hhs[dtm['start-date']] = {}

        if self.deltas is not None:
            for hh in self.hh_data:
                try:
                    delt_hh = self.deltas['hhs'][hh['start-date']]
                    hh['import-net-kwh'] = delt_hh['import-net-kwh']
                    hh['export-net-kwh'] = delt_hh['export-net-kwh']
                    hh['import-gen-kwh'] = delt_hh['import-gen-kwh']
                    hh['msp-kwh'] = delt_hh['msp-kwh']
                    hh['used-kwh'] = delt_hh['used-kwh']
                    hh['msp-kw'] = hh['msp-kwh'] * 2
                except KeyError:
                    pass

    def revolve_to_3rd_party_used(self):
        for hh in self.hh_data:
            hh['msp-kwh'] = hh['used-3rd-party-kwh']
            hh['msp-kw'] = hh['msp-kwh'] * 2
            hh['hist-kwh'] = hh['hist-import-3rd-party-kwh'] - \
                hh['hist-export-3rd-party-kwh']
            hh['hist-kw'] = hh['hist-kwh'] * 2
        self.stream_focus = '3rd-party-used'

    def revolve_to_gen_used(self):
        for hh in self.hh_data:
            hh['msp-kwh'] = hh['used-gen-msp-kwh']
            hh['msp-kw'] = hh['msp-kwh'] * 2
            hh['hist-kwh'] = hh['hist-used-gen-msp-kwh']
            hh['hist-kw'] = hh['hist-kwh'] * 2
        self.stream_focus = 'gen-used'

    def get_data_sources(self, start_date, finish_date, forecast_date=None):
        if forecast_date is None:
            forecast_date = self.forecast_date

        if all(
                (
                    self.start_date == start_date,
                    self.finish_date == finish_date,
                    forecast_date == self.forecast_date)):
            yield self
        else:
            month_start = utc_datetime(start_date.year, start_date.month)
            while month_start <= finish_date:
                month_finish = month_start + relativedelta(months=1) - HH
                chunk_start = hh_max(month_start, start_date)
                chunk_finish = hh_min(month_finish, finish_date)
                if self.stream_focus == 'gen-used' and \
                        self.era is not None and (
                        self.deltas is None or len(self.deltas['hhs']) == 0):
                    era = displaced_era(
                        self.sess, self.caches, self.site, chunk_start,
                        chunk_finish, forecast_date)
                    if era is None:
                        return
                else:
                    era = self.era

                site_ds = SiteSource(
                    self.sess, self.site, chunk_start, chunk_finish,
                    forecast_date, self.caches, era, self.era_maps,
                    self.deltas)
                if self.stream_focus == '3rd-party-used':
                    site_ds.revolve_to_3rd_party_used()
                month_start += relativedelta(months=1)
                yield site_ds


ACTUAL_READ_TYPES = ['N', 'N3', 'C', 'X', 'CP']


class SupplySource(DataSource):
    def __init__(
            self, sess, start_date, finish_date, forecast_date, era, is_import,
            caches, bill=None, era_maps=None, deltas=None):
        DataSource.__init__(
            self, sess, start_date, finish_date, forecast_date, caches,
            era_maps, deltas=deltas)

        self.is_displaced = False
        self.bill = bill
        if self.bill is not None:
            self.bill_start = bill.start_date
            self.bill_finish = bill.finish_date
            self.is_last_bill_gen = \
                not self.bill_finish < self.start_date and not \
                self.bill_finish > self.finish_date

        self.site = None
        self.supply = era.supply
        self.supply_name = self.supply.name
        self.source_code = self.supply.source.code
        self.era = era

        if self.supply.generator_type is None:
            self.generator_type_code = None
        else:
            self.generator_type_code = self.supply.generator_type.code

        if era.supply.gsp_group.code in self.era_map_gsp_groups:
            self.gsp_group_code = self.era_map_gsp_groups[
                era.supply.gsp_group.code]
        else:
            self.gsp_group_code = era.supply.gsp_group.code

        if self.supply.dno.dno_code in self.era_map_dnos:
            self.dno_code = self.era_map_dnos[self.supply.dno.dno_code]
            self.dno = Party.get_dno_by_code(sess, self.dno_code)
        else:
            self.dno = self.supply.dno
            self.dno_code = self.dno.dno_code

        era_map_llfcs = self.era_map_llfcs.get(self.dno_code, {})
        self.is_import = is_import
        if is_import:
            if era.imp_mpan_core in self.era_map_mpan_cores:
                self.mpan_core = self.era_map_mpan_cores[era.imp_mpan_core]
            else:
                self.mpan_core = era.imp_mpan_core

            if era.imp_llfc.code in era_map_llfcs:
                llfc_code = era_map_llfcs[era.imp_llfc.code]
                self.llfc = self.dno.get_llfc_by_code(
                    sess, llfc_code, start_date)
            else:
                self.llfc = era.imp_llfc

            self.sc = era.imp_sc
            self.supplier_account = era.imp_supplier_account

            if era.imp_supplier_contract.id in self.era_map_supplier_contracts:
                self.supplier_contract = Contract.get_supplier_by_id(
                    sess,
                    self.era_map_supplier_contracts[
                        era.imp_supplier_contract.id])
            else:
                self.supplier_contract = era.imp_supplier_contract
        else:
            if era.exp_mpan_core in self.era_map_mpan_cores:
                self.mpan_core = self.era_map_mpan_cores[era.exp_mpan_core]
            else:
                self.mpan_core = era.exp_mpan_core

            if era.exp_llfc is None:
                self.mpan_core = llfc_key = "new_export"
                llfc_key = "new_export"
            else:
                self.mpan_core = era.exp_mpan_core
                llfc_key = era.exp_llfc.code
            if llfc_key in era_map_llfcs:
                llfc_code = era_map_llfcs[llfc_key]
                self.llfc = self.dno.get_llfc_by_code(
                    sess, llfc_code, start_date)
            else:
                self.llfc = era.exp_llfc

            self.sc = 0 if era.exp_sc is None else era.exp_sc
            self.supplier_account = era.exp_supplier_account

            if era.exp_supplier_contract is None:
                sup_key = "new_export"
            else:
                sup_key = era.exp_supplier_contract.id

            if sup_key in self.era_map_supplier_contracts:
                self.supplier_contract = Contract.get_supplier_by_id(
                    sess, self.era_map_supplier_contracts[sup_key])
            else:
                self.supplier_contract = era.exp_supplier_contract

        if era.dc_contract.id in self.era_map_dc_contracts:
            self.dc_contract = Contract.get_dc_by_id(
                sess, self.era_map_dc_contracts[era.dc_contract.id])
        else:
            self.dc_contract = era.dc_contract

        if era.mop_contract.id in self.era_map_mop_contracts:
            self.mop_contract = Contract.get_mop_by_id(
                sess, self.era_map_mop_contracts[era.mop_contract.id])
        else:
            self.mop_contract = era.mop_contract

        if era.cop.code in self.era_map_cops:
            self.cop_code = self.era_map_cops[era.cop.code]
        else:
            self.cop_code = era.cop.code

        self.properties = dict(self.era_map.get('properties_overwritten', {}))
        self.properties.update(era.props)
        self.properties.update(
            self.era_map.get('properties_overwrite', {}))

        self.id = self.mpan_core
        self.msn = era.msn
        self.llfc_code = self.llfc.code
        self.voltage_level = self.llfc.voltage_level
        self.voltage_level_code = self.voltage_level.code
        self.is_substation = self.llfc.is_substation
        self.is_new = False
        self.mtc = self.era.mtc
        self.meter_type = self.mtc.meter_type
        self.meter_type_code = self.meter_type.code

        if era.pc.code in self.era_map_pcs:
            self.pc_code = self.era_map_pcs[era.pc.code]
        else:
            self.pc_code = era.pc.code

        era_ssc_code = None if era.ssc is None else era.ssc.code
        if era_ssc_code in self.era_map_sscs:
            ssc_code = self.era_map_sscs[era_ssc_code]
            if ssc_code is None:
                self.ssc = None
            else:
                self.ssc = Ssc.get_by_code(sess, ssc_code)
        else:
            self.ssc = era.ssc

        self.ssc_code = None if self.ssc is None else self.ssc.code

        self.measurement_type = era.meter_category

        self.consumption_info = ''
        hist_map = {}

        if self.years_back == 0:
            hist_eras = [self.era]
        else:
            hist_eras = sess.query(Era).filter(
                Era.supply == self.supply,
                Era.start_date <= self.history_finish,
                or_(
                    Era.finish_date == null(),
                    Era.finish_date >= self.history_start)).order_by(
                Era.start_date)
            if self.is_import:
                hist_eras = hist_eras.filter(Era.imp_mpan_core != null())
            else:
                hist_eras = hist_eras.filter(Era.exp_mpan_core != null())
            hist_eras = hist_eras.options(
                joinedload(Era.pc), joinedload(Era.channels),
                joinedload(Era.mtc).joinedload(Mtc.meter_type)).all()
            if len(hist_eras) == 0:
                hist_eras = sess.query(Era).filter(
                    Era.supply == self.supply).order_by(Era.start_date)
                if self.is_import:
                    hist_eras = hist_eras.filter(Era.imp_mpan_core != null())
                else:
                    hist_eras = hist_eras.filter(Era.exp_mpan_core != null())
                hist_eras = hist_eras.limit(1).all()

        for i, hist_era in enumerate(hist_eras):
            if self.is_import:
                hist_mpan_core = hist_era.imp_mpan_core
            else:
                hist_mpan_core = hist_era.exp_mpan_core

            if hist_mpan_core is None:
                continue

            if self.history_start > hist_era.start_date:
                chunk_start = self.history_start
            else:
                if i == 0:
                    chunk_start = self.history_start
                else:
                    chunk_start = hist_era.start_date

            chunk_finish = hh_min(self.history_finish, hist_era.finish_date)

            hist_measurement_type = hist_era.meter_category

            if hist_measurement_type == 'amr' and self.era_map.get(
                    'use_amr_hh_data', False):
                hist_measurement_type = 'hh'

            if hist_measurement_type == 'unmetered':

                kwh = hist_era.imp_sc * 60 * 30 / (
                    Datetime(chunk_start.year + 1, 1, 1) -
                    Datetime(chunk_start.year, 1, 1)).total_seconds()

                for tpr in sess.query(Tpr).join(MeasurementRequirement).filter(
                        MeasurementRequirement.ssc == hist_era.ssc):
                    for hist_date in hh_range(
                            self.caches, chunk_start, chunk_finish):
                        if is_tpr(sess, self.caches, tpr.code, hist_date):
                            hist_map[hist_date] = {
                                'msp-kw': kwh * 2, 'msp-kwh': kwh,
                                'hist-kwh': kwh, 'imp-msp-kvarh': 0,
                                'imp-msp-kvar': 0, 'exp-msp-kvarh': 0,
                                'exp-msp-kvar': 0, 'status': 'A',
                                'tpr': tpr.code
                            }

            elif self.bill is None and hist_measurement_type in ('nhh', 'amr'):
                self.consumption_info += _no_bill_nhh(
                    sess, caches, self.supply, chunk_start, chunk_finish,
                    hist_map, forecast_date)
            elif self.bill is not None and hist_measurement_type in (
                    'nhh', 'amr'):
                hhd = {}
                for hh_date in hh_range(
                        self.caches, chunk_start, chunk_finish):
                    hhd[hh_date] = {
                        'status': 'X', 'imp-msp-kvarh': 0,
                        'imp-msp-kvar': 0, 'exp-msp-kvarh': 0,
                        'exp-msp-kvar': 0, 'msp-kw': 0, 'msp-kwh': 0,
                        'hist-kwh': 0}

                tpr_codes = [
                    t[0] for t in sess.query(Tpr.code).
                    join(MeasurementRequirement).filter(
                        MeasurementRequirement.ssc == self.ssc)]

                bills = dict(
                    (b.id, b) for b in sess.query(Bill).filter(
                        Bill.supply == self.supply,
                        Bill.start_date <= chunk_finish,
                        Bill.finish_date >= chunk_start).order_by(
                        Bill.issue_date.desc(), Bill.start_date))
                while True:
                    to_del = None
                    for a, b in combinations(bills.values(), 2):
                        if all(
                                (
                                    a.start_date == b.start_date,
                                    a.finish_date == b.finish_date,
                                    a.kwh == -1 * b.kwh, a.net == -1 * b.net,
                                    a.vat == -1 * b.vat,
                                    a.gross == -1 * b.gross)):
                            to_del = (a.id, b.id)
                            break
                    if to_del is None:
                        break
                    else:
                        for k in to_del:
                            del bills[k]

                prev_type_alias = aliased(ReadType)
                pres_type_alias = aliased(ReadType)
                ct_tz = timezone('Europe/London')
                for bill in bills.values():
                    kws = defaultdict(int)
                    for coefficient, previous_date, previous_value, \
                            previous_type, present_date, present_value, \
                            present_type, tpr_code in sess.query(
                            cast(RegisterRead.coefficient, Float),
                            RegisterRead.previous_date,
                            cast(RegisterRead.previous_value, Float),
                            prev_type_alias.code,
                            RegisterRead.present_date,
                            cast(RegisterRead.present_value, Float),
                            pres_type_alias.code, Tpr.code).join(Tpr).join(
                                prev_type_alias,
                                RegisterRead.previous_type_id ==
                                prev_type_alias.id).join(
                            pres_type_alias,
                            RegisterRead.present_type_id ==
                            pres_type_alias.id).filter(
                            RegisterRead.bill == bill,
                            RegisterRead.units == 0).order_by(
                                RegisterRead.present_date):

                        if tpr_code not in tpr_codes:
                            self._add_problem(
                                "The TPR " + str(tpr_code) +
                                " from the register read does not match any " +
                                "of the TPRs (" + ', '.join(tpr_codes) +
                                ") associated with the MPAN.")

                        if present_date > bill.finish_date:
                            self._add_problem(
                                "There's a read after the end of the bill!")
                        advance = present_value - previous_value
                        if advance < 0:
                            self._add_problem("Clocked?")
                            digits = int(log10(previous_value)) + 1
                            advance = 10 ** digits - previous_value + \
                                present_value

                        kwh = advance * coefficient
                        self.consumption_info += "dumb nhh kwh for " + \
                            tpr_code + " is " + str(kwh) + "\n"

                        kws[tpr_code] += kwh

                    for tpr_code, kwh in kws.items():
                        tpr_dict = _tpr_dict(sess, self.caches, tpr_code)
                        days_of_week = tpr_dict['days-of-week']

                        tz = utc if tpr_dict['is-gmt'] else ct_tz

                        if present_type in ACTUAL_READ_TYPES \
                                and previous_type in ACTUAL_READ_TYPES:
                            status = 'A'
                        else:
                            status = 'E'

                        year_delta = relativedelta(years=self.years_back)
                        hh_part = {}
                        for hh_date in hh_range(
                                self.caches, bill.start_date,
                                bill.finish_date):
                            dt = to_tz(tz, hh_date) + year_delta
                            decimal_hour = dt.hour + dt.minute / 60
                            fractional_month = dt.month * 100 + dt.day
                            for ci in days_of_week[dt.weekday()]:

                                if (
                                        (
                                            ci['start-hour'] <
                                            ci['end-hour'] and
                                            ci['start-hour'] <=
                                            decimal_hour < ci['end-hour']) or
                                        (
                                            ci['start-hour'] >=
                                            ci['end-hour'] and (
                                                ci['start-hour'] <=
                                                decimal_hour or
                                                decimal_hour <
                                                ci['end-hour']))) \
                                        and ci['start-month'] <= \
                                        fractional_month <= ci['end-month']:

                                    hh_part[hh_date] = {'status': status}
                                    break

                        num_hh = len(hh_part)
                        if num_hh > 0:
                            rate = kwh / num_hh
                            for d, h in hh_part.items():
                                if chunk_start <= d <= chunk_finish:
                                    hhd_datum = hhd[d]
                                    hhd_datum['msp-kw'] += rate * 2
                                    hhd_datum['msp-kwh'] += rate
                                    hhd_datum['hist-kwh'] += rate
                                    hhd_datum['tpr'] = tpr_code
                                    if hhd_datum['status'] in ('X', 'A'):
                                        hhd_datum['status'] = h['status']
                        elif kwh > 0:
                            self._add_problem(
                                "For the TPR code " + tpr_code +
                                " the bill says that there are " + str(kwh) +
                                " kWh, but the time of the TPR doesn't cover "
                                "the time between the register reads.")

                hist_map.update(hhd)
            elif hist_measurement_type == 'hh':
                full_channels, hhd = _init_hh_data(
                    sess, caches, hist_era, chunk_start, chunk_finish,
                    is_import)
                if not full_channels:
                    full_channels = False
                hist_map.update(hhd)

            else:
                raise BadRequest("gen type not recognized")

        for d in datum_range(
                sess, self.caches, self.years_back, start_date, finish_date):
            datum = d.copy()
            datum.update(hist_map.get(d['hist-start'], {}))
            self.hh_data.append(datum)

            d_start = d['start-date']
            self.supplier_bill_hhs[d_start] = {}
            self.mop_bill_hhs[d_start] = {}
            self.dc_bill_hhs[d_start] = {}

        if self.deltas is not None:
            site_deltas = self.deltas['site']

            try:
                sup_deltas = self.deltas[self.supply.id]
            except KeyError:
                sup_deltas = self.deltas[self.supply.id] = {}

            for hh in self.hh_data:
                hh_start = hh['start-date']
                if hh_start in sup_deltas:
                    delt = sup_deltas[hh_start]
                elif hh_start in site_deltas:
                    delt = sup_deltas[hh_start] = site_deltas[hh_start]
                    del site_deltas[hh_start]
                else:
                    continue

                hh['msp-kwh'] += delt
                hh['msp-kw'] += delt * 2

    def get_data_sources(self, start_date, finish_date, forecast_date=None):
        if forecast_date is None:
            forecast_date = self.forecast_date

        if all(
                (
                    self.start_date == start_date,
                    self.finish_date == finish_date,
                    forecast_date == self.forecast_date)):
            yield self
        elif self.mpan_core in ("new_export", "new_import"):
            yield SupplySource(
                self.sess, start_date, finish_date, forecast_date, self.era,
                self.is_import, self.caches, self.bill, self.era_maps,
                self.deltas)
        else:
            if self.is_import:
                eras = self.sess.query(Era).filter(
                    Era.supply == self.supply,
                    Era.imp_mpan_core != null(),
                    Era.start_date <= finish_date, or_(
                        Era.finish_date == null(),
                        Era.finish_date >= start_date))
            else:
                eras = self.sess.query(Era).filter(
                    Era.supply == self.supply, Era.exp_mpan_core != null(),
                    Era.start_date <= finish_date, or_(
                        Era.finish_date == null(),
                        Era.finish_date >= start_date))
            eras = eras.options(
                joinedload(Era.ssc),
                joinedload(Era.dc_contract),
                joinedload(Era.mop_contract),
                joinedload(Era.imp_supplier_contract),
                joinedload(Era.exp_supplier_contract),
                joinedload(Era.channels),
                joinedload(Era.imp_llfc).joinedload(Llfc.voltage_level),
                joinedload(Era.exp_llfc).joinedload(Llfc.voltage_level),
                joinedload(Era.cop),
                joinedload(Era.supply).joinedload(Supply.dno),
                joinedload(Era.supply).joinedload(Supply.gsp_group),
                joinedload(Era.supply).joinedload(Supply.source),
                joinedload(Era.mtc).joinedload(Mtc.meter_type),
                joinedload(Era.pc), joinedload(Era.site_eras))

            for era in eras:
                chunk_start = hh_max(era.start_date, start_date)
                chunk_finish = hh_min(era.finish_date, finish_date)

                ds = SupplySource(
                    self.sess, chunk_start, chunk_finish, forecast_date, era,
                    self.is_import, self.caches, self.bill, self.era_maps,
                    self.deltas)
                yield ds


def _find_pair(sess, caches, is_forwards, read_list):
    if len(read_list) < 2:
        return

    if is_forwards:
        back, front = read_list[-2], read_list[-1]
    else:
        back, front = read_list[-1], read_list[-2]

    back_reads = back['reads']
    front_reads = front['reads']
    back_date = back['date']
    front_date = front['date']

    if back['msn'] == front['msn'] and back_reads.keys() == front_reads.keys():
        r = hh_range(caches, back_date, front_date - HH)

        tprs = {}
        for tpr_code, initial_val in back_reads.items():
            end_val = front_reads[tpr_code]
            num_hh = len([d for d in r if is_tpr(sess, caches, tpr_code, d)])

            # Clocked?
            if end_val - initial_val < 0:
                digits = int(log10(initial_val)) + 1
                end_val += 10 ** digits

            kwh = end_val * front['coefficients'][tpr_code] - \
                initial_val * back['coefficients'][tpr_code]

            tprs[tpr_code] = kwh / num_hh if num_hh > 0 else 0

        return {
            'start-date': back_date,
            'tprs': tprs
        }


def _find_hhs(caches, sess, pairs, chunk_start, chunk_finish):
    if len(pairs) == 0:
        pairs.append(
            {
                'start-date': chunk_start,
                'tprs': {'00001': 0}
            })

    # set finish dates
    for i in range(1, len(pairs)):
        pairs[i - 1]['finish-date'] = pairs[i]['start-date'] - HH
    pairs[-1]['finish-date'] = None

    # stretch
    if hh_after(pairs[0]['start-date'], chunk_start):
        pairs[0]['start-date'] = chunk_start

    # chop
    if hh_before(pairs[0]['finish-date'], chunk_start):
        del pairs[0]
    if hh_after(pairs[-1]['start-date'], chunk_finish):
        del pairs[-1]

    # squash
    if hh_before(pairs[0]['start-date'], chunk_start):
        pairs[0]['start-date'] = chunk_start
    if hh_after(pairs[-1]['finish-date'], chunk_finish):
        pairs[-1]['finish-date'] = chunk_finish

    hhs = {}
    for pair in pairs:
        d_range = hh_range(caches, pair['start-date'], pair['finish-date'])
        for tpr_code, kwh in pair['tprs'].items():
            dates = [d for d in d_range if is_tpr(sess, caches, tpr_code, d)]

            for date in dates:
                hhs[date] = {
                    'msp-kw': kwh * 2, 'msp-kwh': kwh, 'hist-kwh': kwh,
                    'imp-msp-kvar': 0, 'imp-msp-kvarh': 0, 'exp-msp-kvar': 0,
                    'exp-msp-kvarh': 0, 'tpr': tpr_code
                }
    return hhs


def _set_status(hhs, read_list, forecast_date):
    THRESHOLD = 31 * 48 * 30 * 60
    rl = [r for r in read_list if r['date'] <= forecast_date]
    for k, v in hhs.items():
        try:
            periods = (abs(r['date'] - k).total_seconds() for r in rl)
            next(p for p in periods if p <= THRESHOLD)
            v['status'] = 'A'
        except StopIteration:
            v['status'] = 'E'


def _read_generator(sess, supply, start, is_forwards, is_prev):
    if is_prev:
        r_typ = RegisterRead.previous_type
        r_dt = RegisterRead.previous_date
        r_vl = RegisterRead.previous_value
    else:
        r_typ = RegisterRead.present_type
        r_dt = RegisterRead.present_date
        r_vl = RegisterRead.present_value

    q = sess.query(RegisterRead).join(Bill).join(BillType).join(r_typ).filter(
        RegisterRead.units == 0, ReadType.code.in_(ACTUAL_READ_TYPES),
        Bill.supply == supply, BillType.code != 'W').options(
            joinedload(RegisterRead.bill))

    if is_forwards:
        q = q.filter(r_dt >= start).order_by(r_dt, RegisterRead.id)
    else:
        q = q.filter(r_dt < start).order_by(r_dt.desc(), RegisterRead.id)

    for offset in count():
        r = q.offset(offset).first()
        if r is None:
            break

        if is_prev:
            dt = r.previous_date
        else:
            dt = r.present_date

        bill = sess.query(Bill).join(BillType).filter(
            Bill.supply == supply, Bill.reads.any(),
            Bill.finish_date >= r.bill.start_date,
            Bill.start_date <= r.bill.finish_date,
            BillType.code != 'W').order_by(
            Bill.issue_date.desc(), BillType.code).first()

        if bill.id != r.bill.id:
            continue

        era = supply.find_era_at(sess, dt)
        if era is None:
            era_coeff = None
        else:
            era_properties = PropDict(
                chellow.utils.url_root + 'eras/' + str(era.id),
                loads(era.properties))
            try:
                era_coeff = float(era_properties['coefficient'])
            except KeyError:
                era_coeff = None

        reads = {}
        coeffs = {}
        for coeff, value, tpr_code in sess.query(
                cast(RegisterRead.coefficient, Float), cast(r_vl, Float),
                Tpr.code).filter(
                RegisterRead.units == 0, RegisterRead.bill == r.bill,
                RegisterRead.msn == r.msn, RegisterRead.tpr_id == Tpr.id,
                r_dt == dt):
            reads[tpr_code] = value
            coeffs[tpr_code] = coeff if era_coeff is None else era_coeff

        yield {
            'date': dt, 'reads': reads, 'coefficients': coeffs, 'msn': r.msn
        }


def _no_bill_nhh(sess, caches, supply, start, finish, hist_map, forecast_date):
    read_list = []
    pairs = []
    read_keys = set()

    for is_forwards in (False, True):
        prev_reads = iter(
            _read_generator(sess, supply, start, is_forwards, True))
        pres_reads = iter(
            _read_generator(sess, supply, start, is_forwards, False))
        if is_forwards:
            read_list.reverse()

        for read in _make_reads(is_forwards, prev_reads, pres_reads):
            read_key = read['date'], read['msn']
            if read_key in read_keys:
                continue
            read_keys.add(read_key)

            read_list.append(read)
            pair = _find_pair(sess, caches, is_forwards, read_list)
            if pair is not None:
                pairs.append(pair)
                if not is_forwards or (
                        is_forwards and read_list[-1]['date'] > finish):
                    break

    consumption_info = 'read list - \n' + dumps(read_list) + "\n"
    hhs = _find_hhs(caches, sess, pairs, start, finish)
    _set_status(hhs, read_list, forecast_date)
    hist_map.update(hhs)
    return consumption_info + 'pairs - \n' + dumps(pairs)


def _make_reads(forwards, prev_reads, pres_reads):
    prev_read = next(prev_reads, None)
    pres_read = next(pres_reads, None)
    while prev_read is not None or pres_read is not None:

        if prev_read is None:
            yield pres_read
            pres_read = next(pres_reads, None)

        elif pres_read is None:
            yield prev_read
            prev_read = next(prev_reads, None)

        else:
            if (forwards and prev_read['date'] < pres_read['date']) or (
                    not forwards and prev_read['date'] >= pres_read['date']):
                yield prev_read
                prev_read = next(prev_reads, None)
            else:
                yield pres_read
                pres_read = next(pres_reads, None)


def _init_hh_data(
        sess, caches, hist_era, chunk_start, chunk_finish, is_import):
    hhd = {}
    full_channels = True
    has_exp_active = False
    has_imp_related_reactive = False
    has_exp_related_reactive = False
    for channel in hist_era.channels:
        if not channel.imp_related \
                and channel.channel_type == 'ACTIVE':
            has_exp_active = True
        if channel.imp_related and channel.channel_type in (
                'REACTIVE_IMP', 'REACTIVE_EXP'):
            has_imp_related_reactive = True
        if not channel.imp_related and channel.channel_type in (
                'REACTIVE_IMP', 'REACTIVE_EXP'):
            has_exp_related_reactive = True

    if has_exp_active and not has_exp_related_reactive \
            and has_imp_related_reactive:
        #  old style
        full_channels = False
        data = iter(sess.execute("""
select sum(cast(coalesce(kwh.value, 0) as double precision)),
sum(cast(coalesce(anti_kwh.value, 0) as double precision)),
max(kwh.status),
sum(cast(coalesce(reactive_imp.value, 0) as double precision)),
sum(cast(coalesce(reactive_exp.value, 0) as double precision)),
hh_datum.start_date
from hh_datum
join channel on (hh_datum.channel_id = channel.id)
left join hh_datum as kwh
on (hh_datum.id = kwh.id and channel.channel_type = 'ACTIVE'
and channel.imp_related = :is_import)
left join hh_datum as anti_kwh
on (hh_datum.id = anti_kwh.id and channel.channel_type = 'ACTIVE'
and channel.imp_related != :is_import)
left join hh_datum as reactive_imp
on (hh_datum.id = reactive_imp.id
and channel.channel_type = 'REACTIVE_IMP'
and channel.imp_related is true)
left join hh_datum as reactive_exp
on (hh_datum.id = reactive_exp.id
and channel.channel_type = 'REACTIVE_EXP'
and channel.imp_related is true)
where channel.era_id = :era_id and hh_datum.start_date >= :start_date
and hh_datum.start_date <= :finish_date
group by hh_datum.start_date
order by hh_datum.start_date
""", params={
            'era_id': hist_era.id,
            'start_date': chunk_start,
            'finish_date': chunk_finish,
            'is_import': is_import}))
        (
            msp_kwh, anti_msp_kwh, status, imp_kvarh, exp_kvarh,
            hist_start) = next(data, (None, None, None, None, None, None))

        for hh_date in hh_range(caches, chunk_start, chunk_finish):
            if hh_date == hist_start:
                hhd[hh_date] = {
                    'status': status,
                    'imp-msp-kvarh': imp_kvarh,
                    'imp-msp-kvar': imp_kvarh * 2,
                    'exp-msp-kvarh': exp_kvarh,
                    'exp-msp-kvar': exp_kvarh * 2,
                    'msp-kw': msp_kwh * 2,
                    'msp-kwh': msp_kwh,
                    'anti-msp-kwh': anti_msp_kwh,
                    'hist-kwh': msp_kwh,
                }
                (
                    msp_kwh, anti_msp_kwh, status, imp_kvarh, exp_kvarh,
                    hist_start) = next(
                        data, (None, None, None, None, None, None))
    else:
        # new style
        print("doing new style")
        data = iter(sess.execute("""
select hh_datum.start_date,
max(kwh.status),
sum(cast(coalesce(kwh.value, 0) as double precision)),
sum(cast(coalesce(reactive_imp.value, 0) as double precision)),
sum(cast(coalesce(reactive_exp.value, 0) as double precision))
from hh_datum
join channel on (hh_datum.channel_id = channel.id)
left join hh_datum as kwh
on (hh_datum.id = kwh.id and channel.channel_type = 'ACTIVE'
and channel.imp_related = :is_import)
left join hh_datum as reactive_imp
on (hh_datum.id = reactive_imp.id
and channel.channel_type = 'REACTIVE_IMP'
and channel.imp_related is true)
left join hh_datum as reactive_exp
on (hh_datum.id = reactive_exp.id
and channel.channel_type = 'REACTIVE_EXP'
and channel.imp_related is true)
where channel.era_id = :era_id and hh_datum.start_date >= :start_date
and hh_datum.start_date <= :finish_date
group by hh_datum.start_date
order by hh_datum.start_date
""", params={
            'era_id': hist_era.id,
            'start_date': chunk_start,
            'finish_date': chunk_finish,
            'is_import': is_import}))

        for (
                hist_start, status, msp_kwh, imp_kvarh,
                exp_kvarh) in data:

            datum = {
                'imp-msp-kvarh': imp_kvarh,
                'imp-msp-kvar': imp_kvarh * 2,
                'exp-msp-kvarh': exp_kvarh,
                'exp-msp-kvar': exp_kvarh * 2
            }
            if msp_kwh is not None:
                datum['status'] = status
                datum['hist-kwh'] = msp_kwh
                datum['msp-kwh'] = msp_kwh
                datum['msp-kw'] = msp_kwh * 2

            hhd[hist_start] = datum
    return full_channels, hhd



from collections import defaultdict
from datetime import timedelta
from itertools import combinations, count
from math import floor, log10
from types import MappingProxyType

import chellow.bank_holidays
from chellow.computer import hh_rate
from chellow.models import (
    BillType, Contract, GBill, GEra, GRateScript, GReadType, GRegisterRead,
    get_non_core_contract_id
)
from chellow.utils import (
    HH, PropDict, get_file_rates, hh_after, hh_before, hh_max, hh_min,
    hh_range, to_ct, utc_datetime, utc_datetime_now
)

from dateutil.relativedelta import relativedelta

from sqlalchemy import Float, cast, or_
from sqlalchemy.sql.expression import null

from werkzeug.exceptions import BadRequest

from zish import dumps, loads


def get_times(sess, caches, start_date, finish_date, forecast_date):
    times_cache = get_g_engine_cache(caches, 'times')
    try:
        s_cache = times_cache[start_date]
    except KeyError:
        s_cache = times_cache[start_date] = {}

    try:
        f_cache = s_cache[finish_date]
    except KeyError:
        f_cache = s_cache[finish_date] = {}

    try:
        return f_cache[forecast_date]
    except KeyError:
        if start_date > finish_date:
            raise BadRequest('The start date is after the finish date.')
        times_dict = defaultdict(int)
        dt = finish_date
        years_back = 0
        while dt > forecast_date:
            dt -= relativedelta(years=1)
            years_back += 1

        times_dict['history-finish'] = dt
        times_dict['history-start'] = dt - (finish_date - start_date)

        times_dict['years-back'] = years_back

        f_cache[forecast_date] = times_dict
        return times_dict


def get_g_engine_cache(caches, name):
    try:
        return caches['g_engine'][name]
    except KeyError:
        caches['g_engine'] = defaultdict(dict)
        return caches['g_engine'][name]


def g_contract_func(caches, contract, func_name):
    try:
        ns = caches['g_engine']['funcs'][contract.id]
    except KeyError:
        try:
            contr_func_cache = caches['g_engine']['funcs']
        except KeyError:
            contr_func_cache = get_g_engine_cache(caches, 'funcs')

        try:
            ns = contr_func_cache[contract.id]
        except KeyError:
            ns = {
                'db_id': contract.id, 'properties': contract.make_properties()}
            exec(contract.charge_script, ns)
            contr_func_cache[contract.id] = ns

    return ns.get(func_name, None)


def forecast_date():
    now = utc_datetime_now()
    return utc_datetime(now.year, now.month, 1)


def get_data_sources(ds, start_date, finish_date, forecast_date=None):

    if forecast_date is None:
        forecast_date = ds.forecast_date

    if ds.start_date == start_date and ds.finish_date == finish_date and \
            forecast_date == ds.forecast_date:
        yield ds
    else:
        for g_era in ds.sess.query(GEra).filter(
                GEra.g_supply == ds.g_supply,
                GEra.start_date <= finish_date, or_(
                    GEra.finish_date == null(),
                    GEra.finish_date >= start_date)):
            chunk_start = hh_max(g_era.start_date, start_date)
            chunk_finish = hh_min(g_era.finish_date, finish_date)

            ds = GDataSource(
                ds.sess, chunk_start, chunk_finish, forecast_date, g_era,
                ds.caches, ds.g_bill)
            yield ds


def datum_range(sess, caches, years_back, start_date, finish_date):
    try:
        return caches['g_engine']['datum'][years_back][start_date][finish_date]
    except KeyError:
        try:
            g_engine_cache = caches['g_engine']
        except KeyError:
            g_engine_cache = caches['g_engine'] = {}

        try:
            d_cache_datum = g_engine_cache['datum']
        except KeyError:
            d_cache_datum = g_engine_cache['datum'] = {}

        try:
            d_cache_years = d_cache_datum[years_back]
        except KeyError:
            d_cache_years = d_cache_datum[years_back] = {}

        try:
            d_cache = d_cache_years[start_date]
        except KeyError:
            d_cache = d_cache_years[start_date] = {}

        datum_list = []
        bank_holidays_id = Contract.get_non_core_by_name(
            sess, 'bank_holidays').id
        for dt in hh_range(caches, start_date, finish_date):
            hist_date = dt - relativedelta(years=years_back)
            ct_dt = to_ct(dt)

            utc_is_month_end = (dt + HH).day == 1 and dt.day != 1
            ct_is_month_end = (ct_dt + HH).day == 1 and ct_dt.day != 1

            utc_decimal_hour = dt.hour + dt.minute / 60
            ct_decimal_hour = ct_dt.hour + ct_dt.minute / 60

            bhs = hh_rate(sess, caches, bank_holidays_id, dt)['bank_holidays']

            bank_holidays = [b[5:] for b in bhs]
            utc_is_bank_holiday = dt.strftime("%m-%d") in bank_holidays
            ct_is_bank_holiday = ct_dt.strftime("%m-%d") in bank_holidays

            datum_list.append(
                MappingProxyType(
                    {
                        'hist_start': hist_date, 'start_date': dt,
                        'ct_day': ct_dt.day, 'utc_month': dt.month,
                        'utc_day': dt.day,
                        'utc_decimal_hour': utc_decimal_hour,
                        'utc_year': dt.year, 'utc_hour': dt.hour,
                        'utc_minute': dt.minute, 'ct_year': ct_dt.year,
                        'ct_month': ct_dt.month,
                        'ct_decimal_hour': ct_decimal_hour,
                        'ct_day_of_week': ct_dt.weekday(),
                        'utc_day_of_week': dt.weekday(),
                        'utc_is_bank_holiday': utc_is_bank_holiday,
                        'ct_is_bank_holiday': ct_is_bank_holiday,
                        'utc_is_month_end': utc_is_month_end,
                        'ct_is_month_end': ct_is_month_end,
                        'status': 'X', 'kwh': 0, 'hist_kwh': 0,
                        'unit_code': 'M3', 'unit_factor': 1,
                        'units_consumed': 0,
                        'correction_factor': 1,
                        'calorific_value': 0,
                        'avg_cv': 0
                    }
                )
            )
        datum_tuple = tuple(datum_list)
        d_cache[finish_date] = datum_tuple
        return datum_tuple


ACTUAL_READ_TYPES = ['A', 'C', 'S']
CORRECTION_FACTOR = 1.02264


def g_rates(sess, caches, g_contract_id, date):
    try:
        return caches['g_engine']['rates'][g_contract_id][date]
    except KeyError:
        try:
            ccache = caches['g_engine']
        except KeyError:
            ccache = caches['g_engine'] = {}

        try:
            rss_cache = ccache['rates']
        except KeyError:
            rss_cache = ccache['rates'] = {}

        try:
            cont_cache = rss_cache[g_contract_id]
        except KeyError:
            cont_cache = rss_cache[g_contract_id] = {}

        try:
            return cont_cache[date]
        except KeyError:
            month_after = date + relativedelta(months=1) + relativedelta(
                days=1)
            month_before = date - relativedelta(months=1) - relativedelta(
                days=1)

            rs = sess.query(GRateScript).filter(
                GRateScript.g_contract_id == g_contract_id,
                GRateScript.start_date <= date, or_(
                    GRateScript.finish_date == null(),
                    GRateScript.finish_date >= date)).first()

            if rs is None:
                rs = sess.query(GRateScript).filter(
                    GRateScript.g_contract_id == g_contract_id).order_by(
                    GRateScript.start_date.desc()).first()
                if date < rs.start_date:
                    cstart = month_before
                    cfinish = min(month_after, rs.start_date - HH)
                else:
                    cstart = max(rs.finish_date + HH, month_before)
                    cfinish = month_after
            else:
                cstart = max(rs.start_date, month_before)
                if rs.finish_date is None:
                    cfinish = month_after
                else:
                    cfinish = min(rs.finish_date, month_after)

            vals = PropDict(
                "the rate script " + chellow.utils.url_root +
                "g_rate_scripts/" + str(rs.id) + " ", loads(rs.script), [])
            for dt in hh_range(caches, cstart, cfinish):
                if dt not in cont_cache:
                    cont_cache[dt] = vals

            return vals


def _read_generator(sess, g_supply, start, is_forwards, is_prev):
    if is_prev:
        r_typ = GRegisterRead.prev_type
        r_dt = GRegisterRead.prev_date
    else:
        r_typ = GRegisterRead.pres_type
        r_dt = GRegisterRead.pres_date

    q = sess.query(GRegisterRead).join(GBill).join(BillType).join(
        r_typ).filter(
        GReadType.code.in_(ACTUAL_READ_TYPES), GBill.g_supply == g_supply,
        BillType.code != 'W')

    if is_forwards:
        q = q.filter(r_dt >= start).order_by(r_dt, GRegisterRead.id)
    else:
        q = q.filter(r_dt < start).order_by(r_dt.desc(), GRegisterRead.id)

    for offset in count():
        r = q.offset(offset).first()
        if r is None:
            break

        if is_prev:
            dt = r.prev_date
            vl = r.prev_value
        else:
            dt = r.pres_date
            vl = r.pres_value

        g_bill = sess.query(GBill).join(BillType).filter(
            GBill.g_supply == g_supply, GBill.g_reads.any(),
            GBill.finish_date >= r.g_bill.start_date,
            GBill.start_date <= r.g_bill.finish_date,
            BillType.code != 'W').order_by(
            GBill.issue_date.desc(), BillType.code,
            GBill.reference.desc()).first()

        if g_bill.id != r.g_bill.id:
            continue

        yield {'date': dt, 'value': vl, 'msn': r.msn}


class GDataSource():
    def __init__(
            self, sess, start_date, finish_date, forecast_date, g_era,
            caches, g_bill):
        self.sess = sess
        self.caches = caches
        self.forecast_date = forecast_date
        self.start_date = start_date
        self.finish_date = finish_date
        self.bill_hhs = {}
        times = get_times(sess, caches, start_date, finish_date, forecast_date)
        self.years_back = times['years-back']
        self.history_start = times['history-start']
        self.history_finish = times['history-finish']

        self.problem = ''
        self.bill = defaultdict(int, {'problem': ''})
        self.hh_data = []
        self.rate_sets = defaultdict(set)

        self.g_bill = g_bill
        if self.g_bill is not None:
            self.g_bill_start = g_bill.start_date
            self.g_bill_finish = g_bill.finish_date
            self.is_last_g_bill_gen = \
                not self.g_bill_finish < self.start_date and not \
                self.g_bill_finish > self.finish_date

        self.g_era = g_era
        self.g_supply = g_era.g_supply
        self.mprn = self.g_supply.mprn
        self.g_exit_zone_code = self.g_supply.g_exit_zone.code
        self.g_ldz_code = self.g_supply.g_exit_zone.g_ldz.code
        self.g_dn_code = self.g_supply.g_exit_zone.g_ldz.g_dn.code
        self.account = g_era.account
        self.g_reading_frequency = g_era.g_reading_frequency
        self.g_reading_frequency_code = self.g_reading_frequency.code
        self.g_contract = g_era.g_contract

        self.consumption_info = ''

        if self.years_back == 0:
            hist_g_eras = [self.g_era]
        else:
            hist_g_eras = sess.query(GEra).filter(
                GEra.g_supply == self.g_supply,
                GEra.start_date <= self.history_finish, or_(
                    GEra.finish_date == null(),
                    GEra.finish_date >= self.history_start)).order_by(
                GEra.start_date).all()
            if len(hist_g_eras) == 0:
                hist_g_eras = sess.query(GEra).filter(
                    GEra.g_supply == self.g_supply).order_by(
                    GEra.start_date).limit(1).all()

        g_cv_id = get_non_core_contract_id('g_cv')
        hist_map = {}

        for i, hist_g_era in enumerate(hist_g_eras):
            if self.history_start > hist_g_era.start_date:
                chunk_start = self.history_start
            else:
                if i == 0:
                    chunk_start = self.history_start
                else:
                    chunk_start = hist_g_era.start_date

            chunk_finish = hh_min(hist_g_era.finish_date, self.history_finish)
            if self.g_bill is None:
                self.consumption_info += _no_bill_kwh(
                    sess, caches, self.g_supply, chunk_start, chunk_finish,
                    hist_g_era, g_cv_id, self.g_ldz_code, hist_map,
                    forecast_date)
            else:
                _bill_kwh(
                    sess, self.caches, self.g_supply, hist_g_era, chunk_start,
                    chunk_finish, g_cv_id, hist_map, self.g_ldz_code)

        for d in datum_range(
                sess, self.caches, self.years_back, start_date, finish_date):
            h = d.copy()
            hist_start = h['hist_start']
            h.update(hist_map.get(hist_start, {}))
            h['kwh'] = h['units_consumed'] * h['unit_factor'] * \
                h['correction_factor'] * h['calorific_value'] / 3.6
            h['kwh_avg'] = h['units_consumed'] * h['unit_factor'] * \
                h['correction_factor'] * h['avg_cv'] / 3.6
            h['ug_rate'] = float(
                get_file_rates(
                    self.caches, 'g_ug',
                    h['start_date'])['ug_gbp_per_kwh'][self.g_exit_zone_code])
            self.hh_data.append(h)
            self.bill_hhs[d['start_date']] = {}


def _no_bill_kwh(
        sess, caches, g_supply, start, finish, g_era, g_cv_id, g_ldz_code,
        hist_map, forecast_date):
    read_list = []
    pairs = []
    read_keys = set()

    for is_forwards in (False, True):
        prev_reads = iter(
            _read_generator(sess, g_supply, start, is_forwards, True))
        pres_reads = iter(
            _read_generator(sess, g_supply, start, is_forwards, False))
        if is_forwards:
            read_list.reverse()

        for read in _make_reads(is_forwards, prev_reads, pres_reads):
            read_key = read['date'], read['msn']
            if read_key in read_keys:
                continue
            read_keys.add(read_key)

            read_list.append(read)
            pair = _find_pair(is_forwards, read_list)
            if pair is not None:
                pairs.append(pair)
                if not is_forwards or (
                        is_forwards and read_list[-1]['date'] > finish):
                    break

    consumption_info = 'read list - \n' + dumps(read_list) + "\n"
    hhs = _find_hhs(
        sess, caches, g_era, pairs, start, finish, g_cv_id, g_ldz_code)
    _set_status(hhs, read_list, forecast_date)
    hist_map.update(hhs)
    return consumption_info + 'pairs - \n' + dumps(pairs)


def _find_pair(is_forwards, read_list):
    if len(read_list) < 2:
        return

    if is_forwards:
        back, front = read_list[-2], read_list[-1]
    else:
        back, front = read_list[-1], read_list[-2]

    back_date, back_value = back['date'], back['value']
    front_date, front_value = front['date'], front['value']

    if back['msn'] == front['msn']:
        units = float(front_value - back_value)
        num_hh = (front_date - back_date).total_seconds() / (30 * 60)

        # Clocked?
        if units < 0:
            digits = int(log10(back_value)) + 1
            units += 10 ** digits

        return {
            'start-date': back_date,
            'units': units / num_hh
        }


def _bill_kwh(
        sess, caches, g_supply, hist_g_era, chunk_start, chunk_finish,
        g_cv_id, hist_map, g_ldz_code):

    cf = float(hist_g_era.correction_factor)
    g_unit = hist_g_era.g_unit
    unit_code, unit_factor = g_unit.code, float(g_unit.factor)

    for hh_date in hh_range(caches, chunk_start, chunk_finish):
        cv, avg_cv = find_cv(sess, caches, g_cv_id, hh_date, g_ldz_code)
        hist_map[hh_date] = {
            'unit_code': unit_code,
            'unit_factor': unit_factor,
            'correction_factor': cf,
            'calorific_value': cv,
            'avg_cv': avg_cv
        }

    g_bills = dict(
        (b.id, b) for b in sess.query(GBill).filter(
            GBill.g_supply == g_supply,
            GBill.start_date <= chunk_finish,
            GBill.finish_date >= chunk_start).order_by(
            GBill.issue_date.desc(), GBill.start_date))
    while True:
        to_del = None
        for a, b in combinations(g_bills.values(), 2):
            if all(
                    (
                        a.start_date == b.start_date,
                        a.finish_date == b.finish_date,
                        a.kwh == -1 * b.kwh, a.net == -1 * b.net,
                        a.vat == -1 * b.vat,
                        a.gross == -1 * b.gross)):
                to_del = (a.id, b.id)
                break
        if to_del is None:
            break
        else:
            for k in to_del:
                del g_bills[k]

    for _, g_bill in sorted(g_bills.items()):
        units_consumed = 0
        for prev_value, pres_value in sess.query(
                cast(GRegisterRead.prev_value, Float),
                cast(GRegisterRead.pres_value, Float)).filter(
                GRegisterRead.g_bill == g_bill):
            units_diff = pres_value - prev_value
            if units_diff < 0:
                total_units = 10 ** len(str(int(prev_value)))
                c_units = total_units - prev_value + pres_value
                if c_units < abs(units_diff):
                    units_diff = c_units

            units_consumed += units_diff

        bill_s = (
            g_bill.finish_date - g_bill.start_date +
            timedelta(minutes=30)).total_seconds()
        hh_units_consumed = units_consumed / (bill_s / (60 * 30))

        block_start = hh_max(g_bill.start_date, chunk_start)
        block_finish = hh_min(g_bill.finish_date, chunk_finish)
        for hh_date in hh_range(caches, block_start, block_finish):
            hist_map[hh_date]['units_consumed'] = hh_units_consumed


def find_cv(sess, caches, g_cv_id, dt, g_ldz_code):
    cvs = chellow.computer.hh_rate(
        sess, caches, g_cv_id, dt)['cvs'][g_ldz_code]
    ct = to_ct(dt)
    try:
        cv_props = cvs[ct.day]
    except KeyError:
        cv_props = sorted(cvs.items())[-1][1]

    cv = float(cv_props['cv'])

    try:
        avg_cv = caches['g_engine']['avg_cvs'][g_ldz_code][ct.year][ct.month]
    except KeyError:
        try:
            gec = caches['g_engine']
        except KeyError:
            gec = caches['g_engine'] = {}

        try:
            avg_cache = gec['avg_cvs']
        except KeyError:
            avg_cache = gec['avg_cvs'] = {}

        try:
            avg_cvs_cache = avg_cache[g_ldz_code]
        except KeyError:
            avg_cvs_cache = avg_cache[g_ldz_code] = {}

        try:
            year_cache = avg_cvs_cache[ct.year]
        except KeyError:
            year_cache = avg_cvs_cache[ct.year] = {}

        try:
            avg_cv = year_cache[ct.month]
        except KeyError:
            cv_list = [float(v['cv']) for v in cvs.values()]
            avg_cv = year_cache[ct.month] = floor(
                (sum(cv_list) / len(cv_list)) * 10) / 10
    return cv, avg_cv


def _find_hhs(
        sess, caches, hist_g_era, pairs, chunk_start, chunk_finish, g_cv_id,
        g_ldz_code):
    hhs = {}
    if len(pairs) == 0:
        pairs.append({'start-date': chunk_start, 'units': 0})

    # set finish dates
    for i in range(1, len(pairs)):
        pairs[i - 1]['finish-date'] = pairs[i]['start-date'] - HH
    pairs[-1]['finish-date'] = None

    # stretch
    if hh_after(pairs[0]['start-date'], chunk_start):
        pairs[0]['start-date'] = chunk_start

    # chop
    if hh_before(pairs[0]['finish-date'], chunk_start):
        del pairs[0]
    if hh_after(pairs[-1]['start-date'], chunk_finish):
        del pairs[-1]

    # squash
    if hh_before(pairs[0]['start-date'], chunk_start):
        pairs[0]['start-date'] = chunk_start
    if hh_after(pairs[-1]['finish-date'], chunk_finish):
        pairs[-1]['finish-date'] = chunk_finish

    cf = float(hist_g_era.correction_factor)
    g_unit = hist_g_era.g_unit
    unit_code, unit_factor = g_unit.code, float(g_unit.factor)
    for pair in pairs:
        units = pair['units']
        for hh_date in hh_range(
                caches, pair['start-date'], pair['finish-date']):
            cv, avg_cv = find_cv(sess, caches, g_cv_id, hh_date, g_ldz_code)

            hhs[hh_date] = {
                'unit_code': unit_code,
                'unit_factor': unit_factor,
                'units_consumed': units,
                'correction_factor': cf,
                'calorific_value': cv,
                'avg_cv': avg_cv
            }
    return hhs


def _make_reads(forwards, prev_reads, pres_reads):
    prev_read = next(prev_reads, None)
    pres_read = next(pres_reads, None)
    while prev_read is not None or pres_read is not None:

        if prev_read is None:
            yield pres_read
            pres_read = next(pres_reads, None)

        elif pres_read is None:
            yield prev_read
            prev_read = next(prev_reads, None)

        else:
            if (forwards and prev_read['date'] < pres_read['date']) or (
                    not forwards and prev_read['date'] >= pres_read['date']):
                yield prev_read
                prev_read = next(prev_reads, None)
            else:
                yield pres_read
                pres_read = next(pres_reads, None)


def _set_status(hhs, read_list, forecast_date):
    THRESHOLD = 31 * 48 * 30 * 60
    rl = [r for r in read_list if r['date'] <= forecast_date]
    for k, v in hhs.items():
        try:
            periods = (abs(r['date'] - k).total_seconds() for r in rl)
            next(p for p in periods if p <= THRESHOLD)
            v['status'] = 'A'
        except StopIteration:
            v['status'] = 'E'



import traceback
import pytz
from datetime import datetime as Datetime
from dateutil.relativedelta import relativedelta
from sqlalchemy import or_, func
from sqlalchemy.sql.expression import null, true
import math
import chellow.dloads
import sys
import os
import threading
from chellow.utils import HH, hh_format, req_int, hh_min, hh_max
from chellow.models import (
    Supply, Era, Source, Site, SiteEra, Bill, RegisterRead, BillType, ReadType,
    HhDatum, Channel, Session)
from flask import request, g
from chellow.views import chellow_redirect
import csv
from zish import dumps


def content(year, supply_id, user):
    f = sess = None
    try:
        sess = Session()
        fname = ['crc', str(year), str(year + 1)]
        if supply_id is None:
            fname.append('all_supplies')
        else:
            fname.append('supply_' + str(supply_id))
        running_name, finished_name = chellow.dloads.make_names(
            '_'.join(fname) + '.csv', user)
        f = open(running_name, mode='w', newline='')
        w = csv.writer(f, lineterminator='\n')

        ACTUAL_READ_TYPES = ['N', 'N3', 'C', 'X', 'CP']
        w.writerow(
            (
                'Chellow Supply Id', 'Report Start', 'Report Finish',
                'MPAN Core', 'Site Id', 'Site Name', 'From', 'To',
                'NHH Breakdown', 'Actual HH Normal Days',
                'Actual AMR Normal Days', 'Actual NHH Normal Days',
                'Actual Unmetered Normal Days', 'Max HH Normal Days',
                'Max AMR Normal Days', 'Max NHH Normal Days',
                'Max Unmetered Normal Days', 'Total Actual Normal Days',
                'Total Max Normal Days', 'Data Type', 'HH kWh', 'AMR kWh',
                'NHH kWh', 'Unmetered kwh', 'HH Filled kWh', 'AMR Filled kWh',
                'Total kWh', 'Note'))

        year_start = Datetime(year, 4, 1, tzinfo=pytz.utc)
        year_finish = year_start + relativedelta(years=1) - HH

        supplies = sess.query(Supply).join(Era).join(Source).filter(
            Source.code.in_(('net', 'gen-net')), Era.imp_mpan_core != null(),
            Era.start_date <= year_finish, or_(
                Era.finish_date == null(),
                Era.finish_date >= year_start)).distinct().order_by(Supply.id)
        if supply_id is not None:
            supply = Supply.get_by_id(sess, supply_id)
            supplies = supplies.filter(Supply.id == supply.id)

        meter_types = ('hh', 'amr', 'nhh', 'unmetered')

        for supply in supplies:
            total_kwh = dict([(mtype, 0) for mtype in meter_types])
            filled_kwh = dict([(mtype, 0) for mtype in ('hh', 'amr')])
            normal_days = dict([(mtype, 0) for mtype in meter_types])
            max_normal_days = dict([(mtype, 0) for mtype in meter_types])

            breakdown = ''
            eras = sess.query(Era).filter(
                Era.supply == supply, Era.start_date <= year_finish, or_(
                    Era.finish_date == null(),
                    Era.finish_date >= year_start)).order_by(
                Era.start_date).all()
            supply_from = hh_max(eras[0].start_date, year_start)
            supply_to = hh_min(eras[-1].finish_date, year_finish)

            for era in eras:

                meter_type = era.meter_category

                period_start = hh_max(era.start_date, year_start)
                period_finish = hh_min(era.finish_date, year_finish)

                max_normal_days[meter_type] += (
                    (period_finish - period_start).total_seconds() +
                    60 * 30) / (60 * 60 * 24)

                mpan_core = era.imp_mpan_core
                site = sess.query(Site).join(SiteEra).filter(
                    SiteEra.is_physical == true(),
                    SiteEra.era_id == era.id).one()

                if meter_type in ('nhh', 'amr'):

                    read_list = []
                    read_keys = {}
                    pairs = []

                    prior_pres_reads = iter(
                        sess.query(RegisterRead).join(Bill).join(BillType)
                        .join(RegisterRead.present_type).filter(
                            RegisterRead.units == 0,
                            ReadType.code.in_(ACTUAL_READ_TYPES),
                            Bill.supply == supply,
                            RegisterRead.present_date < period_start,
                            BillType.code != 'W').order_by(
                            RegisterRead.present_date.desc()))
                    prior_prev_reads = iter(
                        sess.query(RegisterRead).join(Bill).join(BillType)
                        .join(RegisterRead.previous_type).filter(
                            RegisterRead.units == 0,
                            ReadType.code.in_(ACTUAL_READ_TYPES),
                            Bill.supply == supply,
                            RegisterRead.previous_date < period_start,
                            BillType.code != 'W').order_by(
                            RegisterRead.previous_date.desc()))
                    next_pres_reads = iter(
                        sess.query(RegisterRead).join(Bill).join(BillType)
                        .join(RegisterRead.present_type).filter(
                            RegisterRead.units == 0,
                            ReadType.code.in_(ACTUAL_READ_TYPES),
                            Bill.supply == supply,
                            RegisterRead.present_date >= period_start,
                            BillType.code != 'W').order_by(
                            RegisterRead.present_date))
                    next_prev_reads = iter(
                        sess.query(RegisterRead).join(Bill).join(BillType).
                        join(RegisterRead.previous_type).filter(
                            RegisterRead.units == 0,
                            ReadType.code.in_(ACTUAL_READ_TYPES),
                            Bill.supply == supply,
                            RegisterRead.previous_date >= period_start,
                            BillType.code != 'W').order_by(
                            RegisterRead.previous_date))

                    for is_forwards in [False, True]:
                        if is_forwards:
                            pres_reads = next_pres_reads
                            prev_reads = next_prev_reads
                            read_list.reverse()
                        else:
                            pres_reads = prior_pres_reads
                            prev_reads = prior_prev_reads

                        prime_pres_read = None
                        prime_prev_read = None
                        while True:
                            while prime_pres_read is None:
                                try:
                                    pres_read = next(pres_reads)
                                except StopIteration:
                                    break

                                pres_date = pres_read.present_date
                                pres_msn = pres_read.msn
                                read_key = '_'.join([str(pres_date), pres_msn])
                                if read_key in read_keys:
                                    continue

                                pres_bill = sess.query(Bill).join(BillType). \
                                    filter(
                                        Bill.reads.any(),
                                        Bill.supply == supply,
                                        Bill.finish_date >=
                                        pres_read.bill.start_date,
                                        Bill.start_date <=
                                        pres_read.bill.finish_date,
                                        BillType.code != 'W').order_by(
                                        Bill.issue_date.desc(),
                                        BillType.code).first()
                                if pres_bill != pres_read.bill:
                                    continue

                                reads = dict(
                                    (
                                        read.tpr.code,
                                        float(read.present_value) *
                                        float(read.coefficient))
                                    for read in sess.query(RegisterRead).
                                    filter(
                                        RegisterRead.units == 0,
                                        RegisterRead.bill == pres_bill,
                                        RegisterRead.present_date == pres_date,
                                        RegisterRead.msn == pres_msn))

                                prime_pres_read = {
                                    'date': pres_date, 'reads': reads,
                                    'msn': pres_msn}
                                read_keys[read_key] = None
                            while prime_prev_read is None:
                                try:
                                    prev_read = next(prev_reads)
                                except StopIteration:
                                    break

                                prev_date = prev_read.previous_date
                                prev_msn = prev_read.msn
                                read_key = '_'.join([str(prev_date), prev_msn])
                                if read_key in read_keys:
                                    continue
                                prev_bill = sess.query(Bill).join(BillType). \
                                    filter(
                                        Bill.reads.any(),
                                        Bill.supply_id == supply.id,
                                        Bill.finish_date >=
                                        prev_read.bill.start_date,
                                        Bill.start_date <=
                                        prev_read.bill.finish_date,
                                        BillType.code != 'W').order_by(
                                        Bill.issue_date.desc(),
                                        BillType.code).first()
                                if prev_bill != prev_read.bill:
                                    continue

                                reads = dict(
                                    (
                                        read.tpr.code,
                                        float(read.previous_value) *
                                        float(read.coefficient))
                                    for read in sess.query(RegisterRead).
                                    filter(
                                        RegisterRead.units == 0,
                                        RegisterRead.bill_id == prev_bill.id,
                                        RegisterRead.previous_date ==
                                        prev_date,
                                        RegisterRead.msn == prev_msn))

                                prime_prev_read = {
                                    'date': prev_date, 'reads': reads,
                                    'msn': prev_msn}
                                read_keys[read_key] = None

                            if prime_pres_read is None and \
                                    prime_prev_read is None:
                                break
                            elif prime_pres_read is None:
                                read_list.append(prime_prev_read)
                                prime_prev_read = None
                            elif prime_prev_read is None:
                                read_list.append(prime_pres_read)
                                prime_pres_read = None
                            else:
                                if is_forwards:
                                    if prime_pres_read['date'] <= \
                                            prime_prev_read['date']:
                                        read_list.append(prime_pres_read)
                                        prime_pres_read = None
                                    else:
                                        read_list.append(prime_prev_read)
                                        prime_prev_read = None
                                else:
                                    if prime_prev_read['date'] >= \
                                            prime_pres_read['date']:
                                        read_list.append(prime_prev_read)
                                        prime_prev_read = None
                                    else:
                                        read_list.append(prime_pres_read)
                                        prime_pres_read = None

                            if len(read_list) > 1:
                                if is_forwards:
                                    aft_read = read_list[-2]
                                    fore_read = read_list[-1]
                                else:
                                    aft_read = read_list[-1]
                                    fore_read = read_list[-2]

                                if aft_read['msn'] == fore_read['msn'] and \
                                        set(aft_read['reads'].keys()) == \
                                        set(fore_read['reads'].keys()):
                                    pair_start_date = aft_read['date'] + HH
                                    pair_finish_date = fore_read['date']

                                    num_hh = (
                                        (
                                            pair_finish_date + HH -
                                            pair_start_date).total_seconds()
                                        ) / (30 * 60)

                                    tprs = {}
                                    for tpr_code, initial_val in \
                                            aft_read['reads'].items():
                                        end_val = fore_read['reads'][tpr_code]

                                        kwh = end_val - initial_val

                                        if kwh < 0:
                                            digits = int(
                                                math.log10(initial_val)) + 1
                                            kwh = 10 ** digits + kwh

                                        tprs[tpr_code] = kwh / num_hh

                                    pairs.append(
                                        {
                                            'start-date': pair_start_date,
                                            'finish-date': pair_finish_date,
                                            'tprs': tprs})

                                    if len(pairs) > 0 and (
                                            not is_forwards or (
                                                is_forwards and
                                                read_list[-1]['date'] >
                                                period_finish)):
                                        break

                    breakdown += 'read list - \n' + dumps(read_list) + "\n"
                    if len(pairs) == 0:
                        pairs.append(
                            {
                                'start-date': period_start,
                                'finish-date': period_finish,
                                'tprs': {'00001': 0}
                            }
                        )
                    else:
                        for pair in pairs:
                            pair_start = pair['start-date']
                            pair_finish = pair['finish-date']
                            if pair_start >= year_start and \
                                    pair_finish <= year_finish:
                                block_start = hh_max(pair_start, period_start)
                                block_finish = hh_min(
                                    pair_finish, period_finish)

                                if block_start <= block_finish:
                                    normal_days[meter_type] += (
                                        (
                                            block_finish - block_start
                                        ).total_seconds() +
                                        60 * 30) / (60 * 60 * 24)

                    # smooth
                    for i in range(1, len(pairs)):
                        pairs[i - 1]['finish-date'] = pairs[i]['start-date'] \
                            - HH

                    # stretch
                    if pairs[0]['start-date'] > period_start:
                        pairs[0]['start-date'] = period_start

                    if pairs[-1]['finish-date'] < period_finish:
                        pairs[-1]['finish-date'] = period_finish

                    # chop
                    pairs = [
                        pair for pair in pairs
                        if not pair['start-date'] > period_finish and
                        not pair['finish-date'] < period_start]

                    # squash
                    if pairs[0]['start-date'] < period_start:
                        pairs[0]['start-date'] = period_start

                    if pairs[-1]['finish-date'] > period_finish:
                        pairs[-1]['finish-date'] = period_finish

                    for pair in pairs:
                        pair_hhs = (
                            (
                                pair['finish-date'] - pair['start-date']
                            ).total_seconds() + 30 * 60) / (60 * 30)
                        pair['pair_hhs'] = pair_hhs
                        for tpr_code, pair_kwh in pair['tprs'].items():
                            total_kwh[meter_type] += pair_kwh * pair_hhs

                    breakdown += 'pairs - \n' + dumps(pairs)

                elif meter_type == 'hh':
                    period_kwhs = list(
                        float(v[0]) for v in sess.query(HhDatum.value).
                        join(Channel).filter(
                            Channel.imp_related == true(),
                            Channel.channel_type == 'ACTIVE',
                            Channel.era == era,
                            HhDatum.start_date >= period_start,
                            HhDatum.start_date <= period_finish).order_by(
                                HhDatum.id))
                    year_kwhs = list(
                        float(v[0]) for v in sess.query(HhDatum.value).
                        join(Channel).join(Era).filter(
                            Channel.imp_related == true(),
                            Channel.channel_type == 'ACTIVE',
                            Era.supply == supply,
                            HhDatum.start_date >= year_start,
                            HhDatum.start_date <= year_finish).order_by(
                                HhDatum.id))

                    period_sum_kwhs = sum(period_kwhs)
                    year_sum_kwhs = sum(year_kwhs)
                    period_len_kwhs = len(period_kwhs)
                    year_len_kwhs = len(year_kwhs)
                    total_kwh[meter_type] += period_sum_kwhs
                    period_hhs = (
                        period_finish + HH - period_start
                        ).total_seconds() / (60 * 30)
                    if year_len_kwhs > 0:
                        filled_kwh[meter_type] += year_sum_kwhs / \
                            year_len_kwhs * (period_hhs - period_len_kwhs)
                    normal_days[meter_type] += sess.query(
                        func.count(HhDatum.value)).join(Channel). \
                        filter(
                            Channel.imp_related == true(),
                            Channel.channel_type == 'ACTIVE',
                            Channel.era == era,
                            HhDatum.start_date >= period_start,
                            HhDatum.start_date <= period_finish,
                            HhDatum.status == 'A').one()[0] / 48
                elif meter_type == 'unmetered':
                    year_seconds = (
                        year_finish - year_start).total_seconds() + 60 * 30
                    period_seconds = (
                        period_finish - period_start).total_seconds() + 60 * 30

                    total_kwh[meter_type] += era.imp_sc * period_seconds / \
                        year_seconds

                    normal_days[meter_type] += period_seconds / (60 * 60 * 24)

            # for full year 183
            total_normal_days = sum(normal_days.values())
            total_max_normal_days = sum(max_normal_days.values())
            is_normal = total_normal_days / total_max_normal_days >= 183 / 365

            w.writerow(
                [
                    supply.id, hh_format(year_start), hh_format(year_finish),
                    mpan_core, site.code, site.name, hh_format(supply_from),
                    hh_format(supply_to), breakdown] +
                [
                    normal_days[t] for t in meter_types] + [
                    max_normal_days[t] for t in meter_types] + [
                    total_normal_days, total_max_normal_days,
                    "Actual" if is_normal else "Estimated"] +
                [total_kwh[t] for t in meter_types] +
                [filled_kwh[t] for t in ('hh', 'amr')] +
                [sum(total_kwh.values()) + sum(filled_kwh.values()), ''])

            # avoid a long running transaction
            sess.rollback()
    except BaseException:
        msg = traceback.format_exc()
        sys.stderr.write(msg + '\n')
        f.write("Problem " + msg)
    finally:
        if sess is not None:
            sess.close()
        if f is not None:
            f.close()
            os.rename(running_name, finished_name)


def do_get(sess):
    year = req_int('year')
    supply_id = req_int('supply_id') if 'supply_id' in request.values else None
    user = g.user
    threading.Thread(target=content, args=(year, supply_id, user)).start()
    return chellow_redirect("/downloads", 303)



from collections import namedtuple
from datetime import datetime as Datetime
from decimal import Decimal, InvalidOperation
from io import StringIO

from chellow.edi_lib import EdiParser, SEGMENTS
from chellow.models import Session, Ssc, Supply
from chellow.utils import HH, parse_mpan_core, to_ct, to_utc

from dateutil.relativedelta import relativedelta

from werkzeug.exceptions import BadRequest


READ_TYPE_MAP = {
    '00': 'N',
    '01': 'Q',
    '02': 'E',
    '03': 'F',
    '04': 'C',
    '05': 'M',
    '06': 'I',
}

SSC_MAP = {
    # 0326
    '20 0000 5855 970': {
        'Night': '00187',
        'Other': '00210',
        'Weekday': '00184'
    },
    '20 0000 6048 528': {
        'Night': '00187',
        'Other': '00210',
        'Weekday': '00184'
    },
    '20 0002 1909 377': {
        'Night': '00187',
        'Other': '00210',
        'Weekday': '00184'
    },
    '20 0002 2254 214': {
        'Night': '00187',
        'Other': '00210',
        'Weekday': '00184'
    },
    '20 0002 2326 515': {
        'Night': '00187',
        'Other': '00210',
        'Weekday': '00184'
    },
    '20 0002 2371 155': {
        'Night': '00187',
        'Other': '00210',
        'Weekday': '00184'
    },
    '20 0002 5282 171': {
        'Night': '00187',
        'Other': '00210',
        'Weekday': '00184'
    },
    '20 0002 5476 287': {
        'Night': '00187',
        'Other': '00210',
        'Weekday': '00184'
    },
    '20 0002 6228 157': {
        'Night': '00187',
        'Other': '00210',
        'Weekday': '00184'
    },
    '20 0002 6440 184': {
        'Night': '00187',
        'Other': '00210',
        'Weekday': '00184'
    },
    '20 0002 6419 768': {
        'Night': '00187',
        'Other': '00210',
        'Weekday': '00184'
    },
    # 0246
    '22 0001 3834 361': {
        'Night': '00277',
        'Other': '00160',
        'Weekday': '00276',
    },
    '22 0001 4442 321': {
        'Night': '00277',
        'Other': '00160',
        'Weekday': '00276',
    },
    '22 0002 1401 756': {
        'Night': '00277',
        'Other': '00160',
        'Weekday': '00276',
    },
    '22 0002 1427 578': {
        'Night': '00277',
        'Other': '00160',
        'Weekday': '00276',
    },
    '22 0002 1442 103': {
        'Night': '00277',
        'Other': '00160',
        'Weekday': '00276',
    },
    '22 0002 1502 222': {
        'Night': '00277',
        'Other': '00160',
        'Weekday': '00276',
    },
    '22 0002 1528 986': {
        'Night': '00277',
        'Other': '00160',
        'Weekday': '00276',
    },
    '22 0002 1540 176': {
        'Night': '00277',
        'Other': '00160',
        'Weekday': '00276',
    },
    '22 0002 1589 599': {
        'Night': '00277',
        'Other': '00160',
        'Weekday': '00276',
    },
    '22 0002 1823 142': {
        'Night': '00277',
        'Other': '00160',
        'Weekday': '00276',
    },
    '22 0002 1829 376': {
        'Night': '00277',
        'Other': '00160',
        'Weekday': '00276',
    },
    '22 0002 1839 541': {
        'Night': '00277',
        'Other': '00160',
        'Weekday': '00276',
    },
    '22 0002 1878 160': {
        'Night': '00277',
        'Other': '00160',
        'Weekday': '00276',
    },
    '22 0002 1394 727': {
        'Night': '00277',
        'Other': '00160',
        'Weekday': '00276',
    },
    # 0174
    '20 0002 6308 157': {
        'Day': '01072',
        'Night': '01071'
    },
    '20 0002 6467 511': {
        'Day': '01072',
        'Night': '01071'
    },
    '0008': {
        'Other': '00159',
    },
    '0065': {
        'Other': '00010',
    },
    '0036': {
        'Other': '00151',
    },
    '0037': {
        'Other': '00153',
    },
    '0151': {
        'Day': '00043',
        'Night': '00210',
    },
    '0154': {
        'Day': '00039',
        'Night': '00221',
    },
    '0174': {
        'Day': '01071',
        'Night': '01072'
    },
    '0179': {
        'Day': '01139',
        'Night': '01140'
    },
    '0184': {
        'Day': '01149',
        'Night': '01150'
    },
    '0186': {
        'Day': '01153',
        'Night': '01154'
    },
    '0242': {
        'Day': '00044',
        'Night': '00208'
    },
    '0244': {
        'Day': '00040',
        'Night': '00206'
    },
    '0246': {
        'Night': '00160',
        'Other': '00277',
        'Weekday': '00276',
    },
    '0252': {
        'Other': '00212',
    },
    '0265': {
        'Other': '00190',
    },
    '0319': {
        'Other': '00071',
        'Weekday': '00183',
    },
    '0320': {
        'Other': '00072',
        'Weekday': '00184',
    },
    '0322': {
        'Weekday': '01073',
        'Other': '01074',
    },
    '0326': {
        'Night': '00210',
        'Other': '00187',
        'Weekday': '00184',
        'Day': '00184',
    },
    '0393': {
        'Day': '00001',
        'Energy Charges': '00001',
    },
    '0428': {
        'Energy Charges': '00258',
        'Energy Charges 2': '00259',
    },
}


# None denotes a TPR-based charge

TMOD_MAP = {
    '700285': ('standing-gbp', 'standing-rate', 'standing-days'),
    '422733': ('ccl-gbp', 'ccl-rate', 'ccl-kwh'),
    '066540': ('ccl-gbp', 'ccl-rate', 'ccl-kwh'),
    '453043': None,
    '068476': None,
    '265091': None,
    '517180': None,
    '547856': None,
}


BillElement = namedtuple(
    'BillElement', ['gbp', 'rate', 'cons', 'titles', 'desc'])


def _to_date(component):
    return to_utc(to_ct(Datetime.strptime(component, "%y%m%d")))


def _to_decimal(components, divisor=None):
    comp_0 = components[0]
    if comp_0 == '':
        return None

    try:
        result = Decimal(comp_0)
    except InvalidOperation as e:
        raise BadRequest(
            "Can't parse '" + str(comp_0) + "' of " + str(components) +
            " as a decimal:" + str(e))

    if len(components) > 1 and components[-1] == "R":
        result *= Decimal("-1")

    if divisor is not None:
        result /= Decimal(divisor)

    return result


def _find_elements(code, elements):
    segment_name = code + elements[1][0] if code == 'CCD' else code
    elem_codes = [m['code'] for m in SEGMENTS[segment_name]['elements']]
    return dict(zip(elem_codes, elements))


class Parser():
    def __init__(self, f):
        self.parser = EdiParser(
            StringIO(str(f.read(), 'utf-8', errors='ignore')))
        self.line_number = None

    def make_raw_bills(self):
        raw_bills = []
        sess = Session()
        headers = {'sess': sess}
        for self.line_number, code in enumerate(self.parser):
            elements = _find_elements(code, self.parser.elements)
            line = self.parser.line
            try:
                bill = _process_segment(code, elements, line, headers)
            except BadRequest as e:
                raise BadRequest(
                    f"Can't parse the line number {self.line_number} "
                    f"{line} : {e.description}")

            if bill is not None:
                raw_bills.append(bill)
        sess.close()
        return raw_bills


def _process_segment(code, elements, line, headers):
    if 'breakdown' in headers:
        headers['breakdown']['raw-lines'].append(line)

    if code == "BCD":
        headers['issue_date'] = _to_date(elements['IVDT'][0])
        headers['reference'] = elements['INVN'][0]
        headers['bill_type_code'] = elements['BTCD'][0]

        sumo = elements['SUMO']
        headers['start_date'] = _to_date(sumo[0])
        headers['finish_date'] = _to_date(sumo[1]) + relativedelta(days=1) - HH

    elif code == "BTL":
        _process_BTL(elements, headers)

    elif code == "MHD":
        _process_MHD(elements, headers)

    elif code == "CCD":
        _process_CCD(elements, headers)

    elif code == 'CLO':
        _process_CLO(elements, headers)

    elif code == "MTR":
        return _process_MTR(elements, headers)

    elif code == "MAN":
        _process_MAN(elements, headers)


def _process_BTL(elements, headers):
    headers['net'] = Decimal('0.00') + _to_decimal(elements['UVLT'], '100')
    headers['vat'] = Decimal('0.00') + _to_decimal(elements['UTVA'], '100')
    headers['gross'] = Decimal('0.00') + _to_decimal(elements['TBTL'], '100')


def _process_CLO(elements, headers):
    cloc = elements['CLOC']
    headers['account'] = cloc[1]
    # headers['msn'] = cloc[2] if len(cloc) > 2 else ''


def _process_MAN(elements, headers):
    madn = elements['MADN']
    dno = madn[0]
    unique = madn[1]
    check_digit = madn[2]
    # pc = madn[3]
    # mtc = madn[4]
    # llfc = madn[5]

    headers['mpan_core'] = parse_mpan_core(''.join([dno, unique, check_digit]))


def _process_MTR(elements, headers):
    if headers['message_type'] != "UTLBIL":
        return

    sess = headers['sess']
    mpan_core = headers['mpan_core']
    start_date = headers['start_date']
    reads = headers['reads']
    supply = Supply.get_by_mpan_core(sess, mpan_core)
    era = supply.find_era_at(sess, start_date)
    bill_elements = []
    if era is None:
        era = supply.find_last_era(sess)

    if era is None:
        imp_mpan_core = ''
        ssc = Ssc.get_by_code(sess, '0393')
    else:
        imp_mpan_core = era.imp_mpan_core
        ssc = Ssc.get_by_code(sess, '0393') if era.ssc is None else era.ssc

    try:
        ssc_lookup = imp_mpan_core
        tpr_map = SSC_MAP[ssc_lookup]
    except KeyError:
        ssc_lookup = ssc.code
        try:
            tpr_map = SSC_MAP[ssc_lookup]
        except KeyError:
            raise BadRequest(
                f"The SSC {ssc_lookup} isn't in the SSC_MAP.")

    for read in reads:
        desc = read['tpr_code']
        try:
            read['tpr_code'] = tpr_map[desc]
        except KeyError:
            raise BadRequest(
                f"The description {desc} isn't in the SSC_MAP "
                f"for the SSC {ssc_lookup}.")

    for el in headers['bill_elements']:
        if el.desc == 'Energy Charges':
            # If it's an unmetered supply there is only one charge
            # line in the EDI, no mater how many TPRs there are.
            # Therefore we split the charge evenly between the
            # TPRs.
            mrs = ssc.measurement_requirements
            num_mrs = len(mrs)
            gbp = el.gbp / num_mrs
            cons = el.cons / num_mrs
            for mr in mrs:
                tpr_code = mr.tpr.code
                titles = (
                    tpr_code + '-gbp', tpr_code + '-rate', tpr_code + '-kwh')
                bill_elements.append(
                    BillElement(
                        gbp=gbp, rate=el.rate, cons=cons, titles=titles,
                        desc=None))
        else:
            if el.titles is None:
                try:
                    tpr = tpr_map[el.desc]
                except KeyError:
                    raise BadRequest(
                        f"The billing element description {el.desc} "
                        f"isn't in the SSC_MAP for the SSC {ssc_lookup}.")

                titles = (tpr + '-gbp', tpr + '-rate', tpr + '-kwh')
            else:
                titles = el.titles

            bill_elements.append(
                BillElement(
                    gbp=el.gbp, titles=titles, rate=el.rate, cons=el.cons,
                    desc=None))

    breakdown = headers['breakdown']
    for bill_el in bill_elements:
        eln_gbp, eln_rate, eln_cons = bill_el.titles

        try:
            breakdown[eln_gbp] += bill_el.gbp
        except KeyError:
            breakdown[eln_gbp] = bill_el.gbp

        rate = bill_el.rate
        if eln_rate is not None and rate is not None:
            try:
                rates = breakdown[eln_rate]
            except KeyError:
                rates = breakdown[eln_rate] = set()

            rates.add(rate)

        cons = bill_el.cons
        if eln_cons is not None and cons is not None:
            try:
                breakdown[eln_cons] += cons
            except KeyError:
                breakdown[eln_cons] = cons

    return {
        'kwh': headers['kwh'],
        'reference': headers['reference'],
        'mpan_core': mpan_core,
        'issue_date': headers['issue_date'],
        'account': headers['account'],
        'start_date': start_date,
        'finish_date': headers['finish_date'],
        'net': headers['net'],
        'vat': headers['vat'],
        'gross': headers['gross'],
        'breakdown': breakdown,
        'reads': reads,
        'bill_type_code': headers['bill_type_code'],
    }


def _process_MHD(elements, headers):
    message_type = elements['TYPE'][0]
    sess = headers['sess']
    if message_type == "UTLBIL":
        headers.clear()
        headers['kwh'] = Decimal('0')
        headers['reads'] = []
        headers['breakdown'] = {'raw-lines': []}
        headers['bill_elements'] = []
        headers['sess'] = sess
    headers['message_type'] = message_type


def _process_CCD(elements, headers):
    ccde = elements['CCDE']
    consumption_charge_indicator = ccde[0]

    if consumption_charge_indicator == "1":
        _process_CCD_1(elements, headers)

    elif consumption_charge_indicator == "2":
        _process_CCD_2(elements, headers)

    elif consumption_charge_indicator == '3':
        _process_CCD_3(elements, headers)

    elif consumption_charge_indicator == '4':
        tmod_1 = elements['TMOD'][0]
        try:
            eln_gbp, eln_rate, eln_cons = TMOD_MAP[tmod_1]
        except KeyError:
            raise BadRequest(
                "Can't find the Tariff Modifer Code 1 " + tmod_1 +
                " in the TMOD_MAP.")

        '''
        m = elements['MLOC'][0]
        mpan_core = ' '.join((m[:2], m[2:6], m[6:10], m[10:]))
        '''
        breakdown = headers['breakdown']

        cons = elements['CONS']
        if eln_cons is not None and len(cons[0]) > 0:
            el_cons = _to_decimal(cons, '1000')
            breakdown[eln_cons] = el_cons

        if eln_rate is not None:
            rate = _to_decimal(elements['BPRI'], '100000')
            try:
                rates = breakdown[eln_rate]
            except KeyError:
                rates = breakdown[eln_rate] = set()

            rates.append(rate)

        '''
        start_date = _to_date(elements['CSDT'][0])
        finish_date = _to_date(elements['CEDT'][0]) - HH
        '''

        if 'CTOT' in elements:
            net = Decimal('0.00') + _to_decimal(elements['CTOT'], '100')
        else:
            net = Decimal('0.00')

        breakdown[eln_gbp] = net


def _process_CCD_1(elements, headers):
    pres_read_date = _to_date(elements['PRDT'][0]) + relativedelta(days=1) - HH

    prev_read_date = _to_date(elements['PVDT'][0]) + relativedelta(days=1) - HH

    m = elements['MLOC'][0]
    mpan = ' '.join(
        (m[13:15], m[15:18], m[18:], m[:2], m[2:6], m[6:10], m[10:13])
    )

    prrd = elements['PRRD']
    try:
        pres_read_type = READ_TYPE_MAP[prrd[1]]
    except KeyError as e:
        raise BadRequest(
            "The present register read type isn't recognized " + str(e))

    try:
        prev_read_type = READ_TYPE_MAP[prrd[3]]
    except KeyError as e:
        raise BadRequest(
            "The previous register read type isn't recognized " + str(e))

    coefficient = Decimal(elements['ADJF'][1]) / Decimal(100000)
    pres_reading_value = Decimal(prrd[0])
    prev_reading_value = Decimal(prrd[2])
    msn = elements['MTNR'][0]
    tpr_code = elements['TMOD'][0]
    if tpr_code == 'kW':
        units = 'kW'
        tpr_code = None
    elif tpr_code == 'kVA':
        units = 'kVA'
        tpr_code = None
    else:
        if tpr_code == '':
            tpr_code = elements['TCOD'][1]

        units = 'kWh'

    headers['reads'].append(
        {
            'msn': msn,
            'mpan': mpan,
            'coefficient': coefficient,
            'units': units,
            'tpr_code': tpr_code,
            'prev_date': prev_read_date,
            'prev_value': prev_reading_value,
            'prev_type_code': prev_read_type,
            'pres_date': pres_read_date,
            'pres_value': pres_reading_value,
            'pres_type_code': pres_read_type
        }
    )


def _process_CCD_2(elements, headers):
    tmod_1 = elements['TMOD'][0]
    try:
        titles = TMOD_MAP[tmod_1]
    except KeyError:
        raise BadRequest(
            "Can't find the Tariff Modifier Code 1 " + tmod_1 +
            " in the TMOD_MAP.")

    '''
    m = elements['MLOC'][0]
    mpan_core = ' '.join((m[:2], m[2:6], m[6:10], m[10:]))
    '''

    if tmod_1 == '700285':  # standing charge
        start_date = _to_date(elements['CSDT'][0])
        finish_date = _to_date(elements['CEDT'][0])
        elcons = Decimal((finish_date - start_date).days + 1)
    else:
        cons = elements['CONS']
        if len(cons[0]) > 0:
            elcons = _to_decimal(cons, '1000')

    rate = _to_decimal(elements['BPRI'], '100000')

    if 'CTOT' in elements:
        gbp = Decimal('0.00') + _to_decimal(elements['CTOT'], '100')
    else:
        gbp = Decimal('0.00')

    headers['bill_elements'].append(
        BillElement(gbp=gbp, titles=titles, rate=rate, cons=elcons, desc=None))


def _process_CCD_3(elements, headers):
    tcod = elements['TCOD']

    tmod_1 = elements['TMOD'][0]
    try:
        titles = TMOD_MAP[tmod_1]
    except KeyError:
        raise BadRequest(
            "Can't find the Tariff Code Modifier 1 {tmod_1} in the TMOD_MAP.")

    if len(tcod) == 2:
        desc = tcod[1]
    else:
        desc = None

    '''
    m = elements['MLOC'][0]
    mpan_core = ' '.join((m[:2], m[2:6], m[6:10], m[10:]))
    '''
    if tmod_1 == '700285':  # standing charge
        start_date = _to_date(elements['CSDT'][0])
        finish_date = _to_date(elements['CEDT'][0])
        consumption = Decimal((finish_date - start_date).days + 1)
    else:
        cons = elements['CONS']
        if len(cons[0]) > 0:
            consumption = _to_decimal(cons, '1000')
        else:
            consumption = Decimal('0')

    if titles is None:
        headers['kwh'] += consumption

    rate = _to_decimal(elements['BPRI'], '100000')

    gbp = Decimal('0.00')
    if 'CTOT' in elements:
        gbp += _to_decimal(elements['CTOT'], '100')

    headers['bill_elements'].append(
        BillElement(
            gbp=gbp, rate=rate, cons=consumption, titles=titles, desc=desc))



from chellow.utils import get_file_rates


def vb(ds):
    to_exit_commodity_rate_set = ds.rate_sets['to_exit_commodity_rate']
    so_exit_commodity_rate_set = ds.rate_sets['so_exit_commodity_rate']
    dn_system_commodity_rate_set = ds.rate_sets['dn_system_commodity_rate']
    dn_system_capacity_rate_set = ds.rate_sets['dn_system_capacity_rate']
    dn_system_capacity_fixed_rate_set = ds.rate_sets[
        'dn_system_capacity_fixed_rate']
    dn_customer_capacity_rate_set = ds.rate_sets['dn_customer_capacity_rate']
    dn_customer_capacity_fixed_rate_set = ds.rate_sets[
        'dn_customer_capacity_fixed_rate']
    dn_customer_fixed_rate_set = ds.rate_sets['dn_customer_fixed_rate']
    dn_ecn_rate_set = ds.rate_sets['dn_ecn_rate']
    dn_ecn_fixed_rate_set = ds.rate_sets['dn_ecn_fixed_rate']

    for hh in ds.hh_data:
        soq = hh['soq']
        aq = hh['aq']
        nts_rates = get_file_rates(
            ds.caches, 'g_nts_commodity', hh['start_date'])

        to_exit_commodity_rate = float(nts_rates['to_exit_gbp_per_kwh'])
        to_exit_commodity_rate_set.add(to_exit_commodity_rate)
        hh['to_exit_commodity_rate'] = to_exit_commodity_rate

        so_exit_commodity_rate = float(nts_rates['so_exit_gbp_per_kwh'])
        so_exit_commodity_rate_set.add(so_exit_commodity_rate)
        hh['so_exit_commodity_rate'] = so_exit_commodity_rate

        rates = get_file_rates(ds.caches, 'g_dn', hh['start_date'])
        dn_rates = rates['gdn'][ds.g_dn_code]
        system_commodity_rates = dn_rates['system_commodity']
        system_capacity_rates = dn_rates['system_capacity']
        customer_capacity_rates = dn_rates['customer_capacity']
        customer_fixed_rates = dn_rates['customer_fixed']

        if aq <= 73200:
            pref = 'to_73200_'
            dn_system_commodity_rate = float(
                system_commodity_rates[pref + 'gbp_per_kwh'])
            dn_system_capacity_rate = float(
                system_capacity_rates[pref + 'gbp_per_kwh_per_day'])
            dn_customer_capacity_rate = float(
                customer_capacity_rates[pref + 'gbp_per_kwh_per_day'])
            dn_customer_fixed_rate = 0
        elif 73200 < aq < 732000:
            pref = '73200_to_732000_'
            dn_system_commodity_rate = float(
                system_commodity_rates[pref + 'gbp_per_kwh'])
            dn_system_capacity_rate = float(
                system_capacity_rates[pref + 'gbp_per_kwh_per_day'])
            dn_customer_capacity_rate = float(
                customer_capacity_rates[pref + 'gbp_per_kwh_per_day'])
            if ds.g_reading_frequency_code == 'A':
                dn_customer_fixed_rate = float(
                    customer_fixed_rates[pref + "biannual_gbp_per_day"])
            else:
                dn_customer_fixed_rate = float(
                    customer_fixed_rates[pref + "monthly_gbp_per_day"])

        else:
            key = '732000_and_over'

            system_commodity_rts = system_commodity_rates[key]
            dn_system_commodity_rate = max(
                float(system_commodity_rts['minimum_gbp_per_kwh']),
                float(system_commodity_rts['gbp_per_kwh']) * soq **
                float(system_commodity_rts['exponent'])
            )

            system_capacity_rts = system_capacity_rates[key]
            dn_system_capacity_rate = max(
                float(system_capacity_rts['minimum_gbp_per_kwh_per_day']),
                float(system_capacity_rts['gbp_per_kwh_per_day']) * soq **
                float(system_capacity_rts['exponent']))

            customer_capacity_rts = customer_capacity_rates[key]
            dn_customer_capacity_rate = float(
                customer_capacity_rts['gbp_per_kwh_per_day']) * soq ** float(
                customer_capacity_rts['exponent'])

            dn_customer_fixed_rate = 0

        dn_system_commodity_rate_set.add(dn_system_commodity_rate)
        hh['dn_system_commodity_rate'] = dn_system_commodity_rate

        dn_system_capacity_rate_set.add(dn_system_capacity_rate)
        hh['dn_system_capacity_rate'] = dn_system_capacity_rate

        dn_system_capacity_fixed_rate = soq * dn_system_capacity_rate
        dn_system_capacity_fixed_rate_set.add(dn_system_capacity_fixed_rate)
        hh['dn_system_capacity_fixed_rate'] = dn_system_capacity_fixed_rate

        dn_customer_capacity_rate_set.add(dn_customer_capacity_rate)
        hh['dn_customer_capacity_rate'] = dn_customer_capacity_rate

        dn_customer_capacity_fixed_rate = soq * dn_customer_capacity_rate
        dn_customer_capacity_fixed_rate_set.add(
            dn_customer_capacity_fixed_rate)
        hh['dn_customer_capacity_fixed_rate'] = dn_customer_capacity_fixed_rate

        dn_customer_fixed_rate_set.add(dn_customer_fixed_rate)
        hh['dn_customer_fixed_rate'] = dn_customer_fixed_rate

        dn_ecn_rate = float(rates['exit_zones'][ds.g_exit_zone_code][
            'exit_capacity_gbp_per_kwh_per_day'])
        dn_ecn_rate_set.add(dn_ecn_rate)
        hh['dn_ecn_rate'] = dn_ecn_rate

        dn_ecn_fixed_rate = dn_ecn_rate * soq
        dn_ecn_fixed_rate_set.add(dn_ecn_fixed_rate)
        hh['dn_ecn_fixed_rate'] = dn_ecn_fixed_rate



from decimal import Decimal
import decimal
from datetime import datetime as Datetime
from chellow.utils import parse_mpan_core, to_utc, to_ct, ct_datetime
from xlrd import xldate_as_tuple, open_workbook
from werkzeug.exceptions import BadRequest
from chellow.models import Session


def get_ct_date(title_row, row, name, datemode):
    val = get_value(title_row, row, name)
    if isinstance(val, float):
        return to_ct(Datetime(*xldate_as_tuple(val, datemode)))


def get_start_date(title_row, row, name, datemode):
    return to_utc(get_ct_date(title_row, row, name, datemode))


def get_finish_date(title_row, row, name, datemode):
    d = get_ct_date(title_row, row, name, datemode)
    return to_utc(ct_datetime(d.year, d.month, d.day, 23, 30))


def get_value(title_row, row, name):
    idx = None
    name = name.strip().lower()
    for i, title_cell in enumerate(title_row):
        if str(title_cell.value).strip().lower() == name:
            idx = i
            break
    if idx is None:
        raise BadRequest(
            "The name '{name}' can't be found in the titles "
            "{title_row}.".format(name=name, title_row=title_row))
    try:
        return row[idx].value
    except IndexError:
        raise BadRequest(
            "For the row " + str(row) + ", the index is " + str(idx) +
            " which is beyond the end of the row. ")


def get_str(title_row, row, name):
    return get_value(title_row, row, name).strip()


def get_dec(title_row, row, name):
    try:
        return Decimal(str(get_value(title_row, row, name)))
    except decimal.InvalidOperation:
        return None


def get_int(title_row, row, name):
    return int(get_value(title_row, row, name))


METER_RATE = Decimal('60.00')


class Parser():
    def __init__(self, f):
        self.book = open_workbook(file_contents=f.read())
        self.sheet = self.book.sheet_by_index(0)

        self.last_line = None
        self._line_number = None
        self._title_line = None

    @property
    def line_number(self):
        return None if self._line_number is None else self._line_number + 1

    def _set_last_line(self, i, line):
        self._line_number = i
        self.last_line = line
        if i == 0:
            self._title_line = line
        return line

    def make_raw_bills(self):
        row_index = None
        sess = None
        try:
            sess = Session()
            bills = []
            title_row = self.sheet.row(0)
            for row_index in range(1, self.sheet.nrows):
                row = self.sheet.row(row_index)
                val = get_value(title_row, row, 'mpan ref')
                if val is None or val == '':
                    break

                self._set_last_line(row_index, val)
                msn = str(get_value(title_row, row, 'meter')).strip()
                mpan_core = parse_mpan_core(
                    str(get_int(title_row, row, 'mpan ref')))
                start_date = get_start_date(
                    title_row, row, 'start', self.book.datemode)
                issue_date = start_date
                finish_date = get_finish_date(
                    title_row, row, 'end', self.book.datemode)
                check = get_str(title_row, row, 'check')
                if check != 'Billed':
                    continue

                net = METER_RATE / 12
                vat = round(net * Decimal('0.2'), 2)

                breakdown = {
                    'raw_lines': [str(title_row)], 'cop': ['5'],
                    'settlement-status': ['non_settlement'], 'msn': [msn],
                    'meter-rate': [METER_RATE], 'meter-gbp': net,
                }

                bills.append(
                    {
                        'bill_type_code': 'N', 'kwh': Decimal(0), 'vat': vat,
                        'net': net, 'gross': net + vat, 'reads': [],
                        'breakdown': breakdown, 'account': mpan_core,
                        'issue_date': issue_date, 'start_date': start_date,
                        'finish_date': finish_date, 'mpan_core': mpan_core,
                        'reference': '_'.join(
                            (
                                start_date.strftime('%Y%m%d'),
                                finish_date.strftime('%Y%m%d'),
                                issue_date.strftime('%Y%m%d'),
                                mpan_core
                            )
                        )
                    }
                )
                sess.rollback()
        except BadRequest as e:
            raise BadRequest(
                "Row number: " + str(row_index) + " " + e.description)
        finally:
            if sess is not None:
                sess.close()

        return bills
